
require 'sketchup.rb'
require 'extensions.rb'
require 'vector.flat_angle.rb'
#***********************************************************************************************************************************************************************************
#                       EXTRUDE ALONG PATH
#***********************************************************************************************************************************************************************************

#        Name : ExtrudeAlongPath
#
# Description : Creates rectangular faced 'followme' extrusions along a path - 
#		Use it to make walls, fascias, soffits, frames, rails etc,
#		starting from a selection of joined edges
#
#      Author : TIG (c) 7/2005
#		based on an original wall making idea by Didier Bur
#		and using some vertex array ideas from Rick Wilson
#
#       Usage : Select joined lines, arcs, circles, curves, etc.
#		Select "Extrude Along Path" from the Plugins menu. 
#		In the dialog enter Alignment, Width and Height, and then OK.
#
#		The 4 alignments are - 'Edge' [on one of the plan edges], 'Central(Plan)'
#		[centrally on plan], 'Centroid' [centered on the start face] or 'Central(Side)'
#		[centered on a side face of the extrusion]...
#		To extrude below the selected path enter a negative height (e.g. -1000).
#		It is difficult to know in advance which of the two ends of the selected path
#		will be taken as the path's start, so the concept of left/right alignment
#		is not used.  With the 'Edge' and 'Central(Side)' alignments the extrusion goes
#		on to the path's left side, looking in the direction of the path's first vector.
#		To get an extrusion flipped on to other side of the path choose 'No' when asked
#		if the starting face's side is acceptable...  Note that entering a negative
#		width will flip the starting face's default side...
#
#		The start face is then shown highlighted at one end of the path. In a dialog
#		you are asked if the extrusion should start at that end - pick 'Yes' to keep it
#		there or pick 'No' to move it to the other end.  This is useful if a path has 
#		say one level end and one sloping end, since from which end the extrusion starts
#		will affect its form and 'skewing', as with normal 'FollowMe' use...
#		With the 'Edge' and 'Central(Side)' alignments the start face remains highlighted 
#		at the selected path end and a dialog asks for 'Yes' to confirm its side of the 
#		path OR 'No' to flip it on to the other side of the path (right/left).
#
#		A construction line is drawn along the starting edge to show where the extrusion
#		is relative to the selected path - note that its line-style varies depending on 
#		the alignment...
#
#		The extrusion is then made and grouped. The extrusion is grouped so it does not 
#		interact with adjacent surfaces - afterwards just explode it if appropriate. 
#
#		The selected path remains highlighted after the extrusion is made -
#		you can use this to delete or move unwanted paths etc.
#		The path stays selected even after an undo or exit on an error.
#
#		The entered height is measured perpendicular to the the first vector in the path
#		(in the plane of the blue/Z-axis), so with sloping paths it might give unexpected
#		extrusion dimensions - see above about choosing a path end from which to start...
#		A vertical starting path vector is allowed BUT note that the entered 'height'
#		and 'width' will default to the red/X-axis and green/Y-axis respectively...
#
#		The end faces always start and end perpendicular to the path vector as with the
#		standard 'FollowMe' use.  With sloping paths this can give unexpected (but correct) 
#		'skewed' extrusions in the blue/Z-axis...
#		Heights and widths that are less than the length of pieces of the path's edge 
#		will give correct but unexpected extrusions, which need manually tidying...
#		Closed loop paths are fully extruded, but some lines might need manually tidying,
#		this is because like 'FollowMe' sloping paths do not always close neatly and the 
#		extrusion can 'skew' in the blue/Z-axis.  
#		Multiple arc and other complex sloping 3D paths and alignments might give unexpected
#		extrusion results.  
#
#		Heights and widths cannot be zero and will return an error.
#		A branching path returns an error as 'FollowMe' can't decide which path to take.
#		Selected edges that are not joined (i.e. they don't allhave common ends-starts)
#		will not be extruded - only the first edge or joined group of edges will extrude.
#
#
#        Type : tool
#
#     Version : 1.0  30/7/5	first release.
#		1.1  1/8/5	face edged to avoid v5 followme bug etc.
#		1.2  7/8/5	extrusion always perpendicular to path, left/right toggle,
#				vertical starting path now allowed; extrusion grouped
#				to avoid rare coincident-face bug-splats.
#		1.3  17/8/5	path end start choice added, cline stipple changed,
#				rare inside out extrusions corrected, dialogs revamped.
#
###--------------------------------------------------------------------------------
### #################################################################
def extrude_along_path
@error=0 ### 17/8/5 ###
model = Sketchup.active_model
model.start_operation "extrude_along_path"
entities = model.active_entities
ss = model.selection
### show VCB and status info
Sketchup::set_status_text("Extrude Along Path... PARAMETERS...", SB_PROMPT)
Sketchup::set_status_text(" ", SB_VCB_LABEL)
Sketchup::set_status_text(" ", SB_VCB_VALUE)
###
if ss.empty?
  Sketchup::set_status_text("Extrude Along Path... NO SELECTION ! ", SB_PROMPT)
  UI.messagebox("No Selection to Extrude.")
  return nil
end
### #################################################################
def get_vertices
### this next bit is mainly thanks to Rick Wilson's weld.rb ###
	model=Sketchup.active_model
	ents=model.active_entities
	@sel=model.selection
	sl=@sel.length
	verts=[]
	edges=[]
	@newVerts=[]
	startEdge=startVert=nil
#DELETE NON-EDGES, GET THE VERTICES
	@sel.each {|item| edges.push(item) if item.typename=="Edge"}
	edges.each {|edge| verts.push(edge.vertices)}
	verts.flatten!
#FIND AN END VERTEX
	vertsShort=[]
	vertsLong=[]
	vertsEnds=[] ### 17/8/5 ### to ensure array is only ends ###
	verts.each do |v|
		if vertsLong.include?(v)
			vertsShort.push(v)
		else
			vertsLong.push(v)
		end
	end
	vertsLong.each do |v| ### 17/8/5 ### to ensure array is only ends ###
		if not vertsShort.include?(v)
			vertsEnds.push(v)
		end
	end ### 17/8/5 ### to ensure array is only ends ###
	if vertsEnds.length==0 ### i.e. it's looped ###
 		### path start or end ? ### 17/8/5 ###
		if @theEnd==0
		  startVert=vertsLong.first
		  startEdge=startVert.edges.first
		else
		  startVert=vertsLong.last
		  startEdge=startVert.edges.first
		end
		###
		closed=true
	else
		if vertsEnds.length != 2
			Sketchup::set_status_text("Extrude Along Path... PATH ERROR ! ", SB_PROMPT)
			UI.messagebox("The selected Path either branches or is not continuous ! \nCannot make an Extrusion ! \nRe-select a single continuous path... ") ### 17/8/5 ###
			@error=1
			return nil
		else
			### path start or end ? ### 17/8/5 ###
			if @theEnd==0
			  startVert=vertsEnds.first
			else
			  startVert=vertsEnds.last
			end
			###
			closed=false
			startEdge=startVert.edges.first
		end
	end
	@sel.clear
#SORT VERTICES, LIMITING TO THOSE IN THE SELECTION SET
	if startVert==startEdge.start
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Extrude Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	else
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Extrude Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	end
	@newVerts.reverse! if reversed
#CONVERT VERTICES TO POINT3Ds
	@newVerts.collect!{|x| x.position}
### now have an array of vertices in order with NO forced closed loop ...
end ### get_vertices
### near or far ?
@theEnd=0
get_vertices
if @error==1
  return nil
end
### dialog #################################################################
alignments = ["Edge","Central(Plan)","Centroid","Central(Side)"]
enums = [alignments.join("|")]
prompts = ["Alignment: ","Width: ","Height: "]
if not $widthIn
   values = ["Central(Plan)",30.cm,100.cm]
else
   values = [$alignment,$widthIn,$heightIn]
end
results = inputbox prompts, values, enums, "Face Parameters (cm)"
return nil if not results ### i.e. the user cancelled the operation
$alignment,$widthIn,$heightIn = results
### #################################################################
### #################################################################
### restore selection set of edges and display them
def edge_reselector (xnewVerts)
 model = Sketchup.active_model
 ents=model.active_entities
 theEdgeX = []
 0.upto(xnewVerts.length-2) do |i| ### 1/8/5
   theEdgeX[i] = ents.add_line(xnewVerts[i],xnewVerts[i+1])  ### make vertices into edges
 end
 model.selection.clear
 model.selection.add theEdgeX
end
###def
### do main stuff #################################################################
pt1 = @newVerts[0]
pt2 = @newVerts[1]
width = $widthIn
if $widthIn == 0.mm ### can't be 0 ###
   $widthIn = 100.mm
   edge_reselector(@newVerts)
   UI.messagebox("Zero Width NOT allowed ! ")
   return nil
end
###
if $heightIn == 0.mm ### can't be 0 ###
   $heightIn = 1000.mm
   edge_reselector(@newVerts)
   UI.messagebox("Zero Height NOT allowed ! ")
   return nil
end
###
if (pt1.x == pt2.x) and (pt1.y == pt2.y) ### vertical 1st path ###
   vflag = 1
   height = $heightIn
else
   vflag = 0
   height = (($heightIn * (pt1.distance pt2)) / (pt1.distance [pt2.x,pt2.y,pt1.z]))
end
###
if $alignment == "Edge"
   offL = width
   offR = 0
   heightUp = height
   heightDn = 0
   cpatt = "__" ### 17/8/5 ###
end
if $alignment == "Central(Plan)"
   offL = width / 2
   offR = width / 2
   heightUp = height
   heightDn = 0
   cpatt = "." ### 17/8/5 ###
end
if $alignment == "Central(Side)"
   offL = width
   offR = 0
   heightUp = height / 2
   heightDn = height / 2
   cpatt = "_" ### 17/8/5 ###
end
if $alignment == "Centroid"
   offL = width / 2
   offR = width / 2
   heightUp = height / 2
   heightDn = height / 2
   cpatt = "-.-" ### 17/8/5 ###
end
### #################################################################
if vflag == 1 ### vertical ### 7/8/5
  vec = pt1.vector_to [(pt1.x + 1), pt1.y, pt1.z]
else
  vec = pt1.vector_to [pt2.x, pt2.y, pt1.z]
end
### #################################################################
piBy2 = 1.5707963267948965 ### a radian right-angle for calculating vector offset to edges
rotated_vecL = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 + piBy2))) ### 7/8/5
pt1_leftC = pt1.offset(rotated_vecL, offL)
rotated_vecR = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 - piBy2))) ### 7/8/5
pt1_rightC = pt1.offset(rotated_vecR, offR)
if vflag == 1 ### 1st path is vertical ### 7/8/5
  pt1_left  = pt1_leftC.offset([-1,0,0], heightDn)
  pt1_right = pt1_rightC.offset([-1,0,0], heightDn)
  pt2_left  = pt1_leftC.offset([1,0,0], heightUp)
  pt2_right = pt1_rightC.offset([1,0,0], heightUp)
else
  pt1_left  = pt1_leftC.offset([0,0,-1], heightDn)
  pt1_right = pt1_rightC.offset([0,0,-1], heightDn)
  pt2_left  = pt1_leftC.offset([0,0,1], heightUp)
  pt2_right = pt1_rightC.offset([0,0,1], heightUp)
end
###
ents=entities
### #################################################################
### add construction line along first vector ### 7/8/5
cline1 = ents.add_cline(pt1,pt2)
cline1.end=nil
cline1.start=nil
cline1.stipple=cpatt ### 17/8/5 ###
### group the extrusion... ### 7/8/5 #################################################################
group=entities.add_group
entities=group.entities
theFace = entities.add_face(pt1_right, pt1_left, pt2_left, pt2_right) ### 7/8/5
if vflag == 1 and height > 0
   theFace.reverse!
end
if vflag == 1 and width < 0
   theFace.reverse!
end
if vflag == 1 and pt1.z > pt2.z
   theFace.reverse!
end
### check start /end of path ###
keypress = 6 ### = YES
### ask if face starts at correct side ? ### 17/8/5 ###
model.selection.clear
model.selection.add [theFace]
model.selection.add [theFace.edges]
###
Sketchup::set_status_text("Extrude Along Path... END ? ", SB_PROMPT)
keypress = UI.messagebox("Start the Face at this End ? ", MB_YESNO) ###
if keypress != 6
  ### get rid of old stuff first ###
  entities.erase_entities cline1
  entities.erase_entities model.selection
  edge_reselector(@newVerts)
  @theEnd=1
  get_vertices
  ###
  pt1 = @newVerts[0]
  pt2 = @newVerts[1]
  if (pt1.x == pt2.x) and (pt1.y == pt2.y) ### vertical 1st path ###
     vflag = 1
     height = $heightIn
  else
     vflag = 0
     height = (($heightIn * (pt1.distance pt2)) / (pt1.distance [pt2.x,pt2.y,pt1.z]))
  end
  if vflag == 1 ### vertical ### 7/8/5
    vec = pt1.vector_to [(pt1.x + 1), pt1.y, pt1.z]
  else
    vec = pt1.vector_to [pt2.x, pt2.y, pt1.z]
  end
  rotated_vecL = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 + piBy2))) ### 7/8/5
  pt1_leftC = pt1.offset(rotated_vecL, offL)
  rotated_vecR = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 - piBy2))) ### 7/8/5
  pt1_rightC = pt1.offset(rotated_vecR, offR)
  if vflag == 1 ### 1st path is vertical ### 7/8/5
    pt1_left  = pt1_leftC.offset([-1,0,0], heightDn)
    pt1_right = pt1_rightC.offset([-1,0,0], heightDn)
    pt2_left  = pt1_leftC.offset([1,0,0], heightUp)
    pt2_right = pt1_rightC.offset([1,0,0], heightUp)
  else
    pt1_left  = pt1_leftC.offset([0,0,-1], heightDn)
    pt1_right = pt1_rightC.offset([0,0,-1], heightDn)
    pt2_left  = pt1_leftC.offset([0,0,1], heightUp)
    pt2_right = pt1_rightC.offset([0,0,1], heightUp)
  end
  ### add NEW construction line along NEW first vector ### 17/8/5
  cline1 = ents.add_cline(pt1,pt2)
  cline1.end=nil
  cline1.start=nil
  cline1.stipple=cpatt
  ###
  theFace = entities.add_face(pt1_right, pt1_left, pt2_left, pt2_right) ###
if vflag == 1 and height > 0
   theFace.reverse!
end
if vflag == 1 and width < 0
   theFace.reverse!
end
if vflag == 1 and pt1.z > pt2.z
   theFace.reverse!
end
end
### end sorting start / end of path ### 17/8/5 ###
###
keypress = 6 ### = YES
### ask if face starts at correct side ? ### 7/8/5
if $alignment == "Edge" or  $alignment == "Central(Side)"
  model.selection.clear
  model.selection.add [theFace]
  model.selection.add [theFace.edges]
  Sketchup::set_status_text("Extrude Along Path... SIDE ? ", SB_PROMPT)
  keypress = UI.messagebox("Start the Face at this Side ? ", MB_YESNO) ###
end
### if NO then reverse face side ### 7/8/5
if keypress != 6
  entities.erase_entities model.selection
  offR = offR * -1
  offL = offL * -1
  rotated_vecL = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 + piBy2))) ### 7/8/5
  pt1_leftC = pt1.offset(rotated_vecL, offL)
  rotated_vecR = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 - piBy2))) ### 7/8/5
  pt1_rightC = pt1.offset(rotated_vecR, offR)
 if vflag == 1 ### 1st path is vertical ### 7/8/5
  pt1_left  = pt1_leftC.offset([-1,0,0], heightDn)
  pt1_right = pt1_rightC.offset([-1,0,0], heightDn)
  pt2_left  = pt1_leftC.offset([1,0,0], heightUp)
  pt2_right = pt1_rightC.offset([1,0,0], heightUp)
 else
  pt1_left  = pt1_leftC.offset([0,0,-1], heightDn)
  pt1_right = pt1_rightC.offset([0,0,-1], heightDn)
  pt2_left  = pt1_leftC.offset([0,0,1], heightUp)
  pt2_right = pt1_rightC.offset([0,0,1], heightUp)
 end
  theFace = entities.add_face(pt1_right, pt1_left, pt2_left, pt2_right)
  theFace.reverse!
 if vflag == 1 and height < 0
   theFace.reverse!
 end
 if vflag == 1 and width < 0
   theFace.reverse!
 end
end
### #################################################################
@@theEdges= []
  0.upto(@newVerts.length-2) do |i| ### 1/8/5
  @@theEdges[i] = ents.add_line(@newVerts[i],@newVerts[i+1]) ### make vertices into edges
end
### follow me along selected edges
if height > 0
  if width > 0
    theExtrusion=theFace.reverse!.followme @@theEdges ###
  end
  if width < 0
    theExtrusion=theFace.followme @@theEdges ###
  end
end
if height < 0
  if width > 0
    theExtrusion=theFace.followme @@theEdges ###
  end
  if width < 0
    theExtrusion=theFace.reverse!.followme @@theEdges ###
  end
end
###
if not theExtrusion
  Sketchup::set_status_text("Extrude Along Path... EXTRUSION ERROR ! ", SB_PROMPT)
  UI.messagebox("NO Extrusion was made ! \nThe selected Path probably branches ! \nRe-select a single continuous path... ")
  entities.erase_entities theFace.edges
  entities.erase_entities cline1
  edge_reselector(@newVerts)
  return nil
end
###
model.commit_operation
### #################################################################
### restore selection set of edges and display them
edge_reselector(@newVerts)
end ### end def
#***********************************************************************************************************************************************************************************#                       # PIPE ALONG PATH
#***********************************************************************************************************************************************************************************
###---------------------------------------------------------------------
#        Name : PipeAlongPath
#
# Description : Creates circular faced 'followme' pipe extrusions along 
#		a path - 
#		Use it to make pipes, ducts etc, starting from a selection of 
#		joined edges
#
#      Author : TIG (c) 9/2005
#		based on TubeAlongPath.rb which itself was from an original 
#		wall making idea by Didier Bur and using some vertex array 
#		ideas from Rick Wilson
#
#       Usage : 
#		Select joined lines, arcs, circles, curves, etc.
#		Select "Pipe Along Path" from the Plugins menu. 
#		In the dialog choose units (mm/inch), enter outside and inside 
#		diameters, number of segements and then OK.  Setting are 
#		remembered during a session.
#
#		The extrusion is made*.
#		*The extrusion is grouped so it does not interact with adjacent 
#		surfaces - afterwards just explode it if appropriate. Edit it 
#		to intesect with model and tidy up to make Tees etc...
#
#		The pipe also has construction points added at each vertex.  
#		These can be used for snapping, if not wanted 'erase 
#		construction geometry' will remove them globally or just within 
#		a group that you are editing.
#
#		One Undo will remove the construction points, another the Pipe.
#
#		The default for extrusion face segments is 24, the minimum is 3.
#
#		The alignment is 'Centroid' - along pipe's centre line.
#
#		The diameter is always measured square to the vector of the 
#		first path's line.  A diameter that is less than the length of 
#		pieces of path edge might give correct but unexpected 
#		extrusions, which might need manually tidying...
#		Closed loop paths are fully extruded.
#		Multiple arcs and other complex 3D paths might give unexpected 
#		results.  Note that SketchUp can't handle very small faces in 
#		its FollowMe mode - so any Arc bends of 8" radius or less that 
#		have the same radius for the pipe (o/d) applied will cause a 
#		crash and "bug splat" and are trapped out BUT note that similar 
#		radii in 'welded' Curves are NOT easily trappable and so they 
#		will always cause a crash with a "bug-splat", so avoid using 
#		these type of path with small radii bends/pipes - or you can 
#		keep it but use 'scale' as explained below...
#		IF you must have this matching small diameter pipe and arc or 
#		welded elbow bends then to get it to work make the path a 
#		group, scale it by a factor so it's larger (say x10), then 
#		edit it and use this tool within it applying the diameter x10, 
#		after it's all made scale the group back down by x0.1 and 
#		explode it and it'll all be OK.
#
#		Both diameters cannot be zero and will return an error.
#		If one of the diamaters is zero then you get a 'tube' rather 
#		than a pipe.
#		If the diamaters entered are equal the inside one is taken as 
#		zero and you will then get a 'tube' rather than a pipe.
#		If the inside's diameter is greater than the outside's then 
#		they are swapped.
#		A branching path returns an error as 'FollowMe' can't decide 
#		which path to take.
#		Selected edges that are not joined (i.e. they don't all have 
#		common ends-starts) will not be extruded - only the first edge 
#		or joined group of edges will extrude.
#
#        Type : Tool
#
#     Version :
#		1.0  18/9/5	first release.
#		1.1  18/9/5	visual segmenting of arced sections etc addressed.
#		1.2  9/2/6  centrepoints added at vertices of non-looped paths,
#					occasional reversed faces on single 'up' line fixed.
#		1.3  16/2/6 undo of cpoints fixed, diam <8" with matching arc 
#					elbow bend radius bug splat trapped (but not curves).
#		1.4  2/5/6	group name -> OD=xx ID=nn with " or mm as units.
#		1.5  3/5/6	Transposed I/O fixed (sorry!).
#   1.6 12/5/7  @error=0 ### fix
#
###---------------------------------------------------------------------
###
def pipe_along_path
model = Sketchup.active_model
entities = model.active_entities
ss = model.selection
### show VCB and status info
Sketchup::set_status_text("Pipe Along Path... PARAMETERS...", SB_PROMPT)
Sketchup::set_status_text(" ", SB_VCB_LABEL)
Sketchup::set_status_text(" ", SB_VCB_VALUE)
###
if ss.empty?
  Sketchup::set_status_text("Pipe Along Path... NO SELECTION ! ", SB_PROMPT)
  UI.messagebox("No Selection to Extrude Pipe.")
  return nil
end
### trap for small radius arcs
aflag=false
arads=[]
for s in ss
  if s.typename=="Edge"
    if s.curve
      if s.curve.typename=="ArcCurve"
        if s.curve.radius < 200.mm
          aflag=true
          arads=arads.push(s.curve.radius)
        end
      else
        for e in s.curve.edges
          if e.curve.typename=="ArcCurve"
            if e.curve.radius < 200.mm
              aflag=true
              arads=arads.push(e.curve.radius)
            end
          end
        end#for
      end
    end
  end
end#for
### ####################################################################
def get_vertices
### this next bit is mainly thanks to Rick Wilson's weld.rb ###
@error=0 ### 20070512
	model=Sketchup.active_model
	ents=model.active_entities
	@sel=model.selection
	sl=@sel.length
	verts=[]
	edges=[]
	@newVerts=[]
	startEdge=startVert=nil
#DELETE NON-EDGES, GET THE VERTICES
	@sel.each {|item| edges.push(item) if item.typename=="Edge"}
	edges.each {|edge| verts.push(edge.vertices)}
	verts.flatten!
#FIND AN END VERTEX
	vertsShort=[]
	vertsLong=[]
	vertsEnds=[] ### to ensure array is only ends ###
	verts.each do |v|
		if vertsLong.include?(v)
			vertsShort.push(v)
		else
			vertsLong.push(v)
		end
	end
	vertsLong.each do |v| ### to ensure array is only ends ###
		if not vertsShort.include?(v)
			vertsEnds.push(v)
		end
	end ### 17/8/5 ### to ensure array is only ends ###
	if vertsEnds.length==0 ### i.e. it's looped ###
 		### path start or end ?
		if @theEnd==0
		  startVert=vertsLong.first
		  @startVert=$endVert=nil ###1.2
		  startEdge=startVert.edges.first
		else
		  startVert=vertsLong.last
		  @startVert=$endVert=nil ###1.2
		  startEdge=startVert.edges.first
		end
		###
		closed=true
	else
		if vertsEnds.length != 2
			Sketchup::set_status_text("Pipe Along Path... PATH ERROR ! ", SB_PROMPT)
			UI.messagebox("The selected Path either branches or is not continuous ! \nCannot make a Pipe Extrusion ! \nRe-select a single continuous path... ")
			@error=1
			return nil
		else
			### path start or end ?
			if @theEnd==0
			  startVert=vertsEnds.first
		      @startVert=startVert ###1.2
		      @endVert=vertsLong.last ###1.2
			else
			  startVert=vertsEnds.last
  		      @startVert=startVert ###1.2
			  @endVert=vertsLong.first ###1.2
			end
			###
			closed=false
			startEdge=startVert.edges.first
		end
	end
	@sel.clear
#SORT VERTICES, LIMITING TO THOSE IN THE SELECTION SET
	if startVert==startEdge.start
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Pipe Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	else
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Pipe Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	end
	@newVerts.reverse! if reversed
#CONVERT VERTICES TO POINT3Ds
	@newVerts.collect!{|x| x.position}
### now have an array of vertices in order with NO forced closed loop ...
end ### get_vertices
###
@theEnd=0
get_vertices
if @error==1
  return nil
end
### dialog #############################################################
units = ["mm","inch"]
enums = [units.join("|")]
if not $diamOutIn
   values = ["mm", 200.0, 0, 6]
else
   values = [$unitsIn, $diamOutIn, $diamInIn, $psegmentsIn]
end
prompts = ["Units (mm / inch) : ", "Outside Diameter: ", "Inside Diameter: ", "Number of Circle Segments: "]
results = inputbox prompts, values, enums, "Pipe Parameters"
return nil if not results ### i.e. the user cancelled the operation
$unitsIn = results[0]
$diamOutIn = results[1]
$diamInIn = results[2]
$psegmentsIn = results[3]
if $diamOutIn == 0 and $diamInIn == 0 ### can't BOTH be 0 ###
   $diamOutIn = 110.0
   UI.messagebox("Two Zero Diameters NOT allowed ! ")
   return nil
end
if $diamOutIn == $diamInIn ### can't be same so Inner=0 ###
   $diamInIn = 0
   UI.messagebox("Equal Diameters NOT allowed !\nInner Diameter is reset to Zero. ")
end
if $diamInIn > $diamOutIn ### Inner can't be bigger ###
   diamInIntemp = $diamInIn
   $diamInIn = $diamOutIn
   $diamOutIn = diamInIntemp
   UI.messagebox("Inner Diameter NOT allowed to be bigger than Outer!\nInner is reset to be the smaller value. ")
end
if $psegmentsIn < 3
   $psegmentsIn = 24
   UI.messagebox("Fewer than 3 Faces are NOT allowed !\nDefaulting to 24. ")
end
segs = $psegmentsIn
radiusO = $diamOutIn / 2
radiusI = $diamInIn / 2
### - do units stuff - ###
if $unitsIn == "mm"
   radiusI = radiusI.mm
   radiusO = radiusO.mm
   txtO=$diamOutIn.to_s+"mm"
   txtI=$diamInIn.to_s+"mm"
else
   radiusI = radiusI.inch
   radiusO = radiusO.inch
   txtO=$diamOutIn.to_s+"\""
   txtI=$diamInIn.to_s+"\""
end
### check for small arcs and == diameter bug-splatters 1.3
dflag=false
for r in arads
  dflag=true if r.inch == radiusO
end
if aflag and dflag ###1.3
  Sketchup::set_status_text("Pipe Along Path... RADIUS ? ! ", SB_PROMPT)
  UI.messagebox("The Radius of at least one Arc in the Path is < 200mm(~8\") \nand it also matches the Pipe's Outer Radius (#{radiusO}) ! \nThis will cause SketchUp to Bug-Splat and Crash !! \nHowever it CAN be done this way - \nScale up the Path and Diameters by x10, \nuse this Tool and then Scale down the Path and Pipe by x0.1...\n\nExiting...")
  return nil
end
### NOW - do main stuff - ###
Sketchup::set_status_text("Pipe Along Path... MAKING PIPE...", SB_PROMPT)
model.start_operation "Pipe Along Path" ###1.3
pt1 = @newVerts[0]
pt2 = @newVerts[1]
vec = pt1.vector_to pt2
group=entities.add_group
entities=group.entities
if not radiusI == 0.0 ###1.1
  circleI = entities.add_circle pt1, vec, radiusI, segs
  baseI = entities.add_face circleI
end
circleO = entities.add_circle pt1, vec, radiusO, segs
baseO = entities.add_face circleO
if not radiusI == 0.0 ###1.1
   baseI.erase!
end
theFace = baseO
if pt1.x == pt2.x and pt1.y == pt2.y and pt1.z < pt2.z ### flip so not inside out ! ###
  theFace.reverse!
end
entits = model.active_entities ###1.1
@@theEdges= []
0.upto(@newVerts.length-2) do |i|
  if @startVert ###1.2
     theFace.reverse! if @newVerts.length < 3 and theFace.normal == [0,0,-1] and @startVert.position.z < @endVert.position.z 
  end ###1.2
  @@theEdges[i] = entits.add_line(@newVerts[i],@newVerts[i+1])  ### make vertices into edges ###1.1
end
### follow me along selected edges
theExtrusion=theFace.reverse!.followme @@theEdges ###
### add name
group.name="OD="+txtO+" ID="+txtI
###
model.commit_operation
### restore selection set of edges and display them
model.start_operation "Pipe's Cpoints" ###1.3
theEdgeX = []
0.upto(@newVerts.length-2) do |i|
  theEdgeX[i] = entits.add_line(@newVerts[i],@newVerts[i+1])  ### make vertices into edges ###1.1
  entities.add_cpoint @newVerts[i] ###if @startVert ###1.2
  entities.add_cpoint @newVerts[i+1] ###if @startVert ###1.2
  entities.add_cpoint [0,0,0.1.mm] if @newVerts[i]==[0,0,0] or @newVerts[i+1]==[0,0,0] ###if @startVert ###1.2
end
model.commit_operation ###1.3
model.selection.clear
model.selection.add theEdgeX
end ### end def ###
#***********************************************************************************************************************************************************************************
#     COPY ALONG PATH
#***********************************************************************************************************************************************************************************
# Copyright 2005, Rick Wilson 
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Name :          pathCopy
# Description :   Copies a selected group/component along a selected path
# Author :        Rick Wilson
# Usage :         1. Install into the plugins directory (or subfolder - supports organize.rb) or into the 
#                    Plugins/examples directory and manually load from the ruby console "load 'examples/bomb.rb'" 
#                 2. Run "Copy along path" from the Plugins menu(or organized submenu).
# Date :          2005-12-05
# Type :          Tool
# History:
#			2.1 (2005-12-07) - fixed problem caused by reversed edges in curves
#			2.0 (2005-12-05) - planned improvements:
#							Specify distance between copies
#							Specify divisor
#			1.0 (2005-11-29) - first version - copies selected component to each vertex of a curve
# 
require 'vector.flat_angle.rb'
class PathCopy
	@@nCursor = 0
	@@type="Node"
#	@@added_items=[]
	@@count=0
def initialize
	@path=nil
	@group=nil
	@state=0
	if $RW_path_copy_spacing
		@spacing=$RW_path_copy_spacing
	else
		@spacing=120.inch
	end
	if(@@nCursor == 0)
		path = Sketchup.find_support_file "pathcopyCursor.png", "Plugins/CD_ICONS"
		@@nCursor = UI::create_cursor(path,0,30) if path
	end
end
def activate
	if @@type=="Spacing"
		Sketchup::set_status_text("Distance between:", SB_VCB_LABEL)
		Sketchup::set_status_text(@spacing, SB_VCB_VALUE)
	else
		Sketchup::set_status_text("Copy to Vertices", SB_VCB_LABEL)
	end
	puts "activating..."
	sel=Sketchup.active_model.selection
	if sel.length>0
		if sel.first.typename=="Edge"
			if sel.first.curve
				@state=1
				@path=sel.first.curve
				puts "path selected"
			else
				@state=1
				@path=sel.first
				puts "path selected"
			end
		end
	    Sketchup::set_status_text($exStrings.GetString("Select group/component to copy"), SB_PROMPT)
	else
	    Sketchup::set_status_text($exStrings.GetString("Select path"), SB_PROMPT)
	end
end
def onSetCursor()
	cursor = UI::set_cursor(@@nCursor)
end
def onLButtonDown(flags,x,y,view)
	if @state>1
		begin
			Sketchup.active_model.commit_operation
		rescue
			puts "nothing to commit"
		end
		@state=0
		@path=nil
#		@group=nil
	end
	ph=view.pick_helper
	num=ph.do_pick x,y
	item=ph.best_picked
#	puts item
	if item
		if @path==nil
			get_path(item) 
		else
			if item.typename=="ComponentInstance"
				points=@path.vertices
#				Sketchup.active_model.start_operation "pathCopy"
				points.collect!{|p| p.position}
				rotation=[]
				points.each {|e| rotation<<0}
				pathCopyComponent(item,points,rotation)
#				Sketchup.active_model.commit_operation
			elsif item.typename=="Group"
				points=@path.vertices
#				Sketchup.active_model.start_operation "pathCopy"
				points.collect!{|p| p.position}
				rotation=[]
				points.each {|e| rotation<<0}
				pathCopyGroup(item,points,rotation)
#				Sketchup.active_model.commit_operation
			end
		end
	end
end
def get_path(item)
	if item
		if item.typename=="Edge"
			if item.curve
				@state=1
				@path=item.curve
				puts "path selected"
			else
				@state=1
				@path=item
				puts "path selected"
			end
			Sketchup::set_status_text($exStrings.GetString("Select group/component to copy"), SB_PROMPT)
		end
	end
	Sketchup.active_model.selection.clear
	Sketchup.active_model.selection.add(@path.edges) if @path.typename=="Curve"|| @path.typename=="ArcCurve"
	Sketchup.active_model.selection.add(@path) if @path.typename=="Edge"
end
def pathCopyComponent(item,points,rotation)
#	@@added_items=[]
	@@count+=1
	Sketchup.active_model.selection.clear
	model=Sketchup.active_model
	ents=model.active_entities
	puts "Copy component instance along path"
#	points=@path.vertices
#	points.collect!{|p| p.position}
#	model.start_operation "pathCopyComponent #{@@count}"
	model.start_operation "pathCopyComponent"
	0.upto(points.length-1) do |i|
		t=Geom::Transformation.new(points[i])
		tr=Geom::Transformation.rotation(points[i],Geom::Vector3d.new(0,0,1),rotation[i])
#		@@added_items<<ents.add_instance(item.definition,t).transform!(tr)
		ents.add_instance(item.definition,t).transform!(tr)
	end
	model.commit_operation
	@state=2
end
def pathCopyGroup(item,points,rotation)
#	@@added_items=[]
	@@count+=1
	Sketchup.active_model.selection.clear
	model=Sketchup.active_model
	ents=model.active_entities
	puts "Copy group along path"
#	points=@path.vertices
#	points.collect!{|p| p.position}
#	model.start_operation "pathCopyGroup #{@@count}"
	model.start_operation "pathCopyGroup"
	0.upto(points.length-1) do |i|
		t1=Geom::Point3d.new(0,0,0)-item.transformation.origin
		t2=Geom::Transformation.new(points[i])
		tr=Geom::Transformation.rotation(points[i],Geom::Vector3d.new(0,0,1),rotation[i])
		gp=item.copy
		gp.transformation=item.transformation
		gp.transform!(t1)
		gp.transform!(t2)
		gp.transform!(tr)
#		@@added_items<<gp
	end
	model.commit_operation
	@state=2
end
def deactivate(view)
#	begin
#		Sketchup.active_model.commit_operation
#	rescue
#		puts "nothing to commit"
#	end
	view.invalidate if @drawn
end
def onCancel(flag, view)
	self.reset(view)
end
def reset(view)
#	begin
#		Sketchup.active_model.commit_operation
#	rescue
#		puts "nothing to commit"
#	end
	@path=nil
	Sketchup::set_status_text($exStrings.GetString("Select path"), SB_PROMPT)
	if( view )
		view.tooltip = nil
		view.invalidate
	end
	
end
end #class PathCopy
class PathCopySpacing < PathCopy
	@@type="Spacing"
	@@count=0
def onLButtonDown(flags,x,y,view)
#	puts "@state is #{@state}"
	if @state>1
#		begin
#			Sketchup.active_model.commit_operation
#		rescue
#			puts "nothing to commit"
#		end
		@state=0
		@path=nil
		@group=nil
		@@count=0
	end
	ph=view.pick_helper
	num=ph.do_pick x,y
	item=ph.best_picked
#	puts item
	if item
		if @path==nil
			get_path(item)
			@state=1
		else
			@group=item
			if @group.typename=="ComponentInstance"
				points,rotation=getPoints(@path)
#				Sketchup.active_model.start_operation "pathCopy #{@@count}"
#				rotation=[]
#				points.each {|e| rotation<<0}
				pathCopyComponent(item,points,rotation)
#				Sketchup.active_model.commit_operation
			elsif @group.typename=="Group"
				points,rotation=getPoints(@path)
#				Sketchup.active_model.start_operation "pathCopy #{@@count}"
#				rotation=[]
#				points.each {|e| rotation<<0}
				pathCopyGroup(item,points,rotation)
#				Sketchup.active_model.commit_operation
			end
		end
	end
end
def onUserText(text, view)
	begin
		value = text.to_l
	rescue
		# Error parsing the text
		UI.beep
		puts "Cannot convert #{text} to a Length"
		value = nil
		Sketchup::set_status_text @spacing, SB_VCB_VALUE
	end
	return if !value
	@spacing=value
	if @state==2
		@@count+=1
#		Sketchup.active_model.start_operation "erase copies"
#		@@added_items.each {|e| e.erase!}
		Sketchup.undo
#		Sketchup.active_model.start_operation "pathCopyRedo #{@@count}"
		points,rotation=getPoints(@path)
		pathCopyComponent(@group,points,rotation) if @group.typename=="ComponentInstance"
		pathCopyGroup(@group,points,rotation) if @group.typename=="Group"
#		Sketchup.active_model.commit_operation
	end
	$RW_path_copy_spacing=@spacing
end
def getPoints(path)
	puts "Getting points from path"
	runningLength=@spacing
	points=[]
	rotation=[]
#	points<<path.edges.first.start.position
#	rotation<<path.edges.first.line[1].flat_angle
	if path.typename=="Edge"
		points<<path.start.position
		rotation<<path.line[1].flat_angle
		if runningLength>path.length
			runningLength-=path.length
		#	puts "edge is only #{path.length.to_f} long, runningLength is now #{runningLength}"
		elsif runningLength==path.length
		#	puts "edge exactly right, adding point #{path.end.position}"
			points<<path.end.position
			rotation<<path.line[1].flat_angle
			runningLength=@spacing
		else
		#	puts "edge is #{path.length.to_f}, longer than runningLength #{runningLength}."
			vec=path.end.position-path.start.position
			vec.length=runningLength
			points<<path.start.position+vec
			rotation<<path.line[1].flat_angle
			runningLength=@spacing-(path.start.position+vec).distance(path.end.position)
		#	puts "subtracting #{(path.start.position+vec).distance(path.end.position).to_f} from edge, runningLength is now #{runningLength}"
			if runningLength<0
				while runningLength<0
					vec.length=@spacing
					points<<points.last+vec
					rotation<<path.line[1].flat_angle
					runningLength=@spacing-points.last.distance(path.end.position)
				#	puts "remaining edge length is #{points.last.distance(path.end.position).to_f}"
				#	puts "runningLength #{runningLength}."
				#	return if UI.messagebox("Continue?",MB_YESNO)==7
				end
			end
		end
	else
		points<<path.edges.first.start.position
		rotation<<path.edges.first.line[1].flat_angle
		path.edges.each do |edge|
 #			puts edge
			if runningLength>edge.length
				runningLength-=edge.length
			#	puts "edge is only #{edge.length.to_f} long, runningLength is now #{runningLength}"
			elsif runningLength==edge.length
			#	puts "edge exactly right, adding point #{edge.end.position}"
				points<<edge.end.position
				rotation<<edge.line[1].flat_angle
				runningLength=@spacing
			else
			#	puts "edge is #{edge.length.to_f}, longer than runningLength #{runningLength}."
				vec=edge.end.position-edge.start.position
				vec.length=runningLength
				points<<edge.start.position+vec
				rotation<<edge.line[1].flat_angle
				runningLength=@spacing-(edge.start.position+vec).distance(edge.end.position)
			#	puts "subtracting #{(edge.start.position+vec).distance(edge.end.position).to_f} from edge, runningLength is now #{runningLength}"
				if runningLength<0
					while runningLength<0
						vec.length=@spacing
						points<<points.last+vec
						rotation<<edge.line[1].flat_angle
						runningLength=@spacing-points.last.distance(edge.end.position)
					#	puts "remaining edge length is #{points.last.distance(edge.end.position).to_f}"
					#	puts "runningLength #{runningLength}."
					#	return if UI.messagebox("Continue?",MB_YESNO)==7
					end
				end
			end
		end
#		puts points
	end
	return points,rotation
end
end #class PathCopySpacing
#***********************************************************************************************************************************************************************************
#                       MIRROR
#***********************************************************************************************************************************************************************************
# Name :          Mirror 3.0 ###
# Description :   Mirrors Selection at Point, Line or Plane
# Author :        Originally by Frank Wiesner
#				  BUT NOTE from v2.6 it's been TIG tweaked and rewritten...###
# Usage :         1. Make Selection (Any Entites allowed)###
#                 2. Select 'Mirror Selection' from Plugins Menu 
#					 or from Context Menu or your ShortCut Key ###
#                 3. If here's no Selection then dialog informs ###
#                 4. Pick a Point and RETURN to Mirror at the Point
#                 5. Pick a Second Point and RETURN to Mirror at the Line
#                 6. Pick a Third Point to Mirror at the Plane
#                 7. Final Option is Erase Original ? ###
#					 If Yes then Highlighting of Selection passes to Copies
#					 (Note: Erase is undo-able separately from the Mirroring) ###
# Type :      	Script
# History:      
#				3.0  05 Aug  2006 ### TIG Ending rewritten to avoid rare bugsplats.
#				2.9  03 Aug  2006 ### TIG Start:commit loops to avoid rare bugsplats.
#				2.8  26 July 2006 ### TIG re-tweaked - Rewritten to be MUCH simpler.
#										Mirroring of ANY Entity Selection works properly.
#				2.7  23 July 2006 ### TIG tweaked - Rewritten generally.
#										Now processes Multiple selections of Groups and/or 
#										Components, but warns if any 'loose' (UnGrouped) 
#										Faces in Selection as they might lose any RePostioned 
#										Texture data, interfere etc.  
#				2.6  22 July 2006 ### TIG tweaked - Context menu added IF ONE group/compo.  
#                                       Warns about selection IF NOT ONE group/compo.  
#                                       Materials now preserved in mirrored copy BUT you will 
#									    need to group selection to preserve material positioning 
#										- although this is define-able it is not find-able from 
#										the orignal face info !  
#                                       Final option to Erase of Original Group/Compo.
#                 2.5  (8.Jul.2oo5)  - bugfix: faces with holes reduce now to a single face
#                                    - major code rewrite (dropped @clondedObject hash, heavy 
#                                      use of groups during construction,...
#                                    - more redundant constructions, but seems to be more robust
#                                    - WARNING: POLYGONS DO NOT WORK VERY WELL
#                 2.4  (5.Jul.2oo5)  - all unnecessary edges created by the add_faces_from_mesh() 
#                                      method are erased! faces with holes mirror now perfectly.
#                 2.3  (4.Jul.2oo5)  - fixed "mirror at plane" error
#                 2.2  (2.Jul.2oo5)  - works in group editting mode
#                                    - prevents picking of identical points
#                 2.1  (1.Jul.2oo5)  - fixed "undo group" bug
#                                    - code cleanup
#                 2.0  (30.Jun.2oo5) - new tool-like interface
#                 1.4  (29.Jun.2oo5) - arc bug fixed
#                                    - code simplified (clone_selection() removed)
#                 1.3  (24.Jun.2oo5) - can handle faces with holes in it (although more edges 
#                                      are drawn than nessesary)
#                                    - correct orientation of mirrored faces 
#                                    - supports arcs (at least works for circles), curves and groups
#                 1.2  (20.Jun.2oo5) - axis and origin selection algorithm does not rely on tool tips any more.
#                 1.1  (19.Jun.2oo5) - fix coplanar point bug
#                                    - better tooltips
#                                    - supports mirror at lines (edges, axis, construction lines)
#                                    - mirror-at-plane code simplified
#                 1.o  (10.Jun.2oo5) - first version
class MirrorTool
    def initialize
        @MIRROR_AT_POINT = 1
        @MIRROR_AT_LINE = 2
        @MIRROR_AT_PLANE = 3
        @numPrecision = 0.00001
    end
    def reset
        @pts = []
        @state = 0
        @transformationType = nil
        @trans = nil
        @point = nil
        @line = nil
        @linev = nil
        @plane = nil
        @planeNormal = nil
        @pointOnPlane = nil
        @curvesDone = Hash.new()
        @cloneGroup = nil
        @ip = Sketchup::InputPoint.new
        @ip1 = Sketchup::InputPoint.new
        @ip2 = Sketchup::InputPoint.new
        @ip3 = Sketchup::InputPoint.new
        @ip = Sketchup::InputPoint.new
        Sketchup::set_status_text "Mirror:  Pick First Point"###
        @drawn = false
    end
    def activate
        ###Sketchup.send_action("showRubyPanel:")
        self.reset
        ss=Sketchup.active_model.selection
        if ss.empty?
            Sketchup::set_status_text "Mirror: NO Selection !"###
    		UI.messagebox("Select Something BEFORE Using the Mirror Tool.")###v2.8
            Sketchup.active_model.select_tool nil
        end
    end
    def deactivate(view)
        view.invalidate if @drawn
        @ip1 = nil
        @ip2 = nil
        @ip3 = nil
    end
    def onMouseMove(flags, x, y, view)
        case @state
        when 0 # getting the first end point
            @ip.pick view, x, y
            if( @ip.valid? && @ip != @ip1 )
                @ip1.copy! @ip
            end
            view.invalidate
            view.tooltip = @ip.tooltip if @ip.valid?
        when 1 # getting the second end point
            @ip.pick view, x, y, @ip1
            if( @ip.valid? && @ip != @ip2 )
                @ip2.copy! @ip
                @pts[1] = @ip2.position
            end
            view.invalidate
            view.tooltip = @ip.tooltip if @ip.valid?
        when 2 # getting the third end point
            @ip.pick view, x, y, @ip2
            if( @ip.valid? && @ip != @ip3 )
                @ip3.copy! @ip
                @pts[1] = @ip3.position
            end
            view.invalidate
            view.tooltip = @ip.tooltip if @ip.valid?
        end
    end
    def onLButtonDown(flags, x, y, view)
        @ip.pick view, x, y
        if( @ip.valid? )
            case @state
            when 0
                if( @ip.valid? )
                    @pts[0] = @ip.position
                    Sketchup::set_status_text "Mirror:  Hit RETURN to Mirror at Point or Pick Point 2"
                    @state = 1
                end
            when 1
                if( @ip.valid? && @ip != @ip1 )
                    @pts[1] = @ip.position
                    @state = 2
                    Sketchup::set_status_text "Mirror:  Hit RETURN to Mirror at Line or Pick Point 3"
                end
            when 2
                if( @ip.valid? && @ip != @ip1 && @ip != @ip2 )
                    @pts[2] = @ip.position
                    @state = 3
                    mirror()
                end
            end
        end
    end
    def onCancel(flag, view)
        view.invalidate if @drawn
        reset
    end
    def draw(view)
        if( @ip.valid? && @ip.display? )
            @ip.draw(view)
            @drawn = true
        end
        if( @state == 1 )
            view.set_color_from_line(@ip1, @ip)
            view.draw(GL_LINE_STRIP, @ip1.position, @ip.position)
            @drawn = true
        elsif( @state == 2 )
            view.drawing_color = "gray"
            view.draw(GL_LINE_STRIP, @ip1.position, @ip2.position)
            view.set_color_from_line(@ip2, @ip)
            view.draw(GL_LINE_STRIP, @ip2.position, @ip.position)
            @drawn = true
        end
    end
    def onKeyDown(key, repeat, flags, view)
        if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
            @shift_down_time = Time.now
            if( view.inference_locked? )
                view.lock_inference
            elsif( @ip.valid? )
                view.lock_inference @ip
            end
        end
    end
    def onKeyUp(key, repeat, flags, view)
        if( key == CONSTRAIN_MODIFIER_KEY && view.inference_locked? && (Time.now - @shift_down_time) > 0.5 )
            view.lock_inference
        end
        if( key == 13)
            mirror()
            self.reset
        end
    end
    
    def mirror()
        case @state
        when 1
            @transformationType = @MIRROR_AT_POINT
            @point = @ip1.position.clone
            @trans = Geom::Transformation.scaling(@point, -1.0)
            Sketchup.active_model.start_operation "Mirror at Point"
        when 2
            @transformationType = @MIRROR_AT_LINE
            @line = [@ip1.position, @ip2.position] 
            @linev = Geom::Vector3d.new(@ip2.position.x-@ip1.position.x, @ip2.position.y-@ip1.position.y, @ip2.position.z-@ip1.position.z)
            Sketchup.active_model.start_operation "Mirror at Line"
        when 3
            @transformationType = @MIRROR_AT_PLANE
            @plane = Geom.fit_plane_to_points(@ip1.position, @ip2.position, @ip3.position) 
            @planeNormal = Geom::Vector3d.new(@plane[0], @plane[1], @plane[2])
            @planeNormal.normalize!
            @pointOnPlane = @ip1.position
            Sketchup.active_model.start_operation "Mirror at Plane"
        end
        ###
        ss=Sketchup.active_model.selection###
    	sents=[]
    	ss.each{|e|sents.push(e)}
    	copy_group=Sketchup.active_model.active_entities.add_group(sents)
        new_group=copy_group.copy
        nents=[]
        new_group.entities.each{|e|nents.push(e)}
        new_center = new_group.bounds.center.clone
        mirror_point(new_center)
        t=nil
        t = Geom::Transformation.scaling(new_center,-1,-1,-1)if @state==1
        t = Geom::Transformation.rotation(new_center,@linev,Math::PI)if @state==2
        tt=nil
        if @state==3
           xxx=@planeNormal.normalize.to_a[0].abs*-1
           yyy=@planeNormal.normalize.to_a[1].abs*-1
           zzz=@planeNormal.normalize.to_a[2].abs*-1
           xxx=1 if xxx>0
           yyy=1 if yyy>0
           zzz=1 if zzz>0
           xxx=-1 if xxx<=0 
           yyy=-1 if yyy<=0
           zzz=-1 if zzz<=0
           t = Geom::Transformation.scaling(new_center,xxx,yyy,zzz)
           tt= Geom::Transformation.rotation(new_center,@planeNormal,Math::PI)
        end#if
        new_group.transform!(t)if t###mirror
        new_group.transform!(tt)if tt###
        transVec = Geom::Vector3d.new(new_center.x-new_group.bounds.center.x, new_center.y-new_group.bounds.center.y, new_center.z-new_group.bounds.center.z) 
        t = Geom::Transformation.translation(transVec)
        new_group.transform!(t)###then move
        ### ending dialog...
#    	Sketchup::set_status_text "Mirror: Erase Original Selection ?"###
#    	if UI.messagebox("Erase Original Selection ? ",MB_YESNO,"")==6
#    	   ### 6=YES 7=NO
#    	   Sketchup.active_model.commit_operation###v2.9
#      	   Sketchup.active_model.start_operation "Mirror: Erase Original Selection"
#    	    Sketchup.active_model.selection.clear
#      	    copy_group.erase!
#    	    new_group.explode
#      	    nents.each{|e|Sketchup.active_model.selection.add(e)if e.valid?}###v3.0
#    	   Sketchup.active_model.commit_operation
#      	else
           Sketchup.active_model.selection.clear
           new_group.explode
           copy_group.explode
           sents.each{|e|Sketchup.active_model.selection.add(e)if e.valid?}###v3.0
           Sketchup.active_model.commit_operation###v2.9
#    	end#if
    	###
        self.reset
        Sketchup.send_action "selectSelectionTool:"
    end#def
    
    def mirror_point(new_p)
        if (@transformationType == @MIRROR_AT_PLANE)
            if (!new_p.on_plane?(@plane))
                mirrorp = new_p.project_to_plane(@plane)
                @trans = Geom::Transformation.scaling(mirrorp, -1.0)
                new_p.transform!(@trans)
            end
        elsif (@transformationType == @MIRROR_AT_LINE)
            if (!new_p.on_line?(@line))
                mirrorp = new_p.project_to_line(@line)
                @trans = Geom::Transformation.scaling(mirrorp, -1.0)
                new_p.transform!(@trans)
            end
        elsif (@transformationType == @MIRROR_AT_POINT)
            if (!(new_p == @point))
                new_p.transform!(@trans)
            end
        end
    end#def
    
end # class MirrorTool
#***********************************************************************************************************************************************************************************
#                       EXTRUDE LINES
#***********************************************************************************************************************************************************************************
# Copyright 2004, D. Bur
# This software is provided as an example of using the Ruby interface
# to SketchUp.
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name : Extrude line Tool
# Description : Extrude lines of selection along a vector given by 2 points
# Menu Item : Plugins / Extrude lines
# Date : 01 Nov 2007
# Type : Tool
#----------------------------------------------------------------------------
class ExtrudeLineTool
def initialize
    $ss = Sketchup.active_model.selection
    $ip1 = nil
    $ip2 = nil
    @xdown = 0
    @ydown = 0
end
def activate
    $ip1 = Sketchup::InputPoint.new
    $ip2 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false
    self.reset(nil)
end
def deactivate(view)
    view.invalidate if @drawn
end
def onMouseMove(flags, x, y, view)
    if( @state == 0 )
        @ip.pick view, x, y
        if( @ip != $ip1 )
            view.invalidate if( @ip.display? or $ip1.display? )
            $ip1.copy! @ip
            view.tooltip = $ip1.tooltip
        end
    else
        $ip2.pick view, x, y, $ip1
        view.tooltip = $ip2.tooltip if( $ip2.valid? )
        view.invalidate
        if( $ip2.valid? )
            length = $ip1.position.distance($ip2.position)
            Sketchup::set_status_text length.to_s, SB_VCB_VALUE
        end
        if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
            @dragging = true
        end
    end
end
def onLButtonDown(flags, x, y, view)
    if( @state == 0 )
        $ip1.pick view, x, y
        if( $ip1.valid? )
            @state = 1
            Sketchup::set_status_text "Second point", SB_PROMPT
            @xdown = x
            @ydown = y
        end
    else
        if( $ip2.valid? )
            self.create_geometry($ip1.position, $ip2.position,view)
            self.reset(view)
        end
    end
    view.lock_inference
end
# The onLButtonUp method is called when the user releases the left mouse button.
def onLButtonUp(flags, x, y, view)
    if( @dragging && $ip2.valid? )
        self.create_geometry($ip1.position, $ip2.position,view)
        self.reset(view)
    end
end
def onKeyDown(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        if( view.inference_locked? )
            view.lock_inference
        elsif( @state == 0 && $ip1.valid? )
            view.lock_inference $ip1
        elsif( @state == 1 && $ip2.valid? )
            view.lock_inference $ip2, $ip1
        end
    end
end
def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end
def onUserText(text, view)
    return if not @state == 1
    return if not $ip2.valid?
    begin
        value = text.to_l
    rescue
        # Error parsing the text
        puts "Cannot convert #{text} to a length."
        value = nil
        Sketchup::set_status_text "", SB_VCB_VALUE
    end
    return if !value
    pt1 = $ip1.position
    vec = $ip2.position - pt1
    if( vec.length == 0.0 )
        return
    end
    vec.length = value
    #pt2 = pt1 + vec
    pt2=pt1.offset vec
    #puts (pt1.distance pt2).to_s
    self.create_geometry(pt1, pt2, view)
    self.reset(view)
end
def draw(view)
    if( $ip1.valid? )
        if( $ip1.display? )
            $ip1.draw(view)
            @drawn = true
            view.draw_points $ip1.position, 8, 1, "black"
        end
        if( $ip2.valid? )
            $ip2.draw(view) if( $ip2.display? )
            view.set_color_from_line($ip1, $ip2)
            self.draw_geometry($ip1.position, $ip2.position, view)
            @drawn = true
        end
    end
    view.draw_points $ip2.position, 8, 1, "black"
end
# onCancel is called when the user hits the escape key
def onCancel(flag, view)
    self.reset(view)
end
# Reset the tool back to its initial state
def reset(view)
    @state = 0
    Sketchup::set_status_text("First point", SB_PROMPT)
    $ip1.clear
    $ip2.clear
    if( view )
        view.tooltip = nil
        view.invalidate if @drawn
    end
    @drawn = false
    @dragging = false
end
# Create new geometry when the user has selected two points.
def create_geometry(p1, p2, view)
model = Sketchup.active_model
model.start_operation "Extrude lines"
entities = model.active_entities
ss = model.selection
indice_line = 0
if ss.empty? 
  UI.messagebox("No selection.")
  return nil
end
# Selection error checking
others = 0
i = 0
0.upto(ss.length - 1) do |something|
  element = ss[i]
  if( element.typename != "Edge")
    others = others + 1
  end
  i = i + 1
end  #of upto
if( others != 0 )
  UI.messagebox( others.to_s + " selected objects are not lines and will be ignored.")
end
dx = p2.x - p1.x
dy = p2.y - p1.y
dz = p2.z - p1.z
i = 0
0.upto(ss.length - 1) do |something|
  if( ss[i].typename == "Edge" )
    pts = ss[i].vertices
    v1 = pts[0].position
    v2 = pts[1].position
    v1_top = Geom::Point3d.new(v1.x+dx, v1.y+dy, v1.z+dz)
    v2_top = Geom::Point3d.new(v2.x+dx, v2.y+dy, v2.z+dz)
    new_face = Sketchup.active_model.active_entities.add_face(v1, v1_top, v2_top, v2)
    if ss[i].soft?
      new_face.edges.each do |e|
        e.smooth = true
        e.soft = true
      end
    end
  end
  #Next line
    i = i + 1
end # of upto
end
# Draw the geometry
def draw_geometry(pt1, pt2, view)
    view.draw_line(pt1, pt2)
end
end # class ExtrudeLineTool
#-----------------------------------------------------------------------------
# This functions is just a shortcut for selecting the new tool
def ex_linetool
   if Sketchup.active_model.selection.empty?
      UI.messagebox("No selection.\nPlease select lines.")
      return
    end
    Sketchup.active_model.select_tool ExtrudeLineTool.new
end
#***********************************************************************************************************************************************************************************
#                       DROP
#***********************************************************************************************************************************************************************************
# Name :          Drop 1.5
# Description :   drop group/component on selected level (change in Y)
# Author :        TBD
# Usage :         1. select groups and/or components, right click, drop at intersection
#                   or
#                 1. select face, right click, get position to set level
#                 2. select groups and/or components, right click, drop at ...
# Date :          22.Jul.2oo4
# Type :          tool
# History:        1.5 (22.Jul.2oo4) - drop selected items at intersection instead of level
#                 1.4 (20.Jul.2oo4) - added debug construction line for drop at intersection
#                                   - bugfixed get position
#                 1.3 (19.Jul.2oo4) - added "find intersection"
#                 1.2 (14.Jul.2oo4) - added support for units
#                 1.1 (14.Jul.2oo4) - added "Drop at ..."
#                 1.o (14.Jul.2oo4) - first version
class Drop
   # Y value for the drop, convert to Length
   @@level = 0.m
   
   # validate the level entity (menu) - allow only Face entities
   def Drop::validate_level_entity
      ss = Sketchup.active_model.selection
      return nil if ss.empty? 
      return true if (ss.first.kind_of? Sketchup::Face and ss.count==1)
   end
   # validate the drop entity (menu)  - allow only Groups and Components 
   def Drop::validate_drop_entity
      ss = Sketchup.active_model.selection
      return nil if ss.empty?
      is_valid = nil
      ss.each do |e| 
         if (e.kind_of? Sketchup::Group) or (e.kind_of? Sketchup::ComponentInstance)   
            is_valid = true 
         else 
            is_valid = false
         end
      end
      return is_valid
   end
   
   # perform the drop
   def Drop::perform_drop (entity)
      position = entity.transformation.to_a
      # Y value at position 14, the rest are orientation and some view values (not sure)
      position[14] = @@level
      entity.transformation = position
      # invalidate the view to see the move action
      Sketchup.active_model.active_view.invalidate
   end
   # get the level from user and perform drop
   def Drop::drop_at
      prompts = ["enter drop level: "]
      values = [@@level]
      results = inputbox prompts, values, "Drop level"
      return if not results 
      @@level = results[0]
      ss = Sketchup.active_model.selection
      ss.each do |e|
         Drop.perform_drop e
      end
   end
   # drop at intersection
   def Drop::drop_intersection
      dir = [0,0,-1]
      ss = Sketchup.active_model.selection
     
      ss.each do |e|
         position = e.transformation.origin
         rt = Sketchup.active_model.raytest(position,dir)
         dir_rt = [position.x, position.y, -1]       
         if rt == nil
            Sketchup.active_model.active_entities.add_cline position,dir_rt
         else
            pt2 = rt[0] 
            @@level = pt2[2]
            Drop.perform_drop e
         end
      end
   end
   
   # get face Y position for drop level
   def Drop::get_plane_position
      ss = Sketchup.active_model.selection
      
      if ss.first.kind_of? Sketchup::Face
         @@level = ss.first.vertices.first.position[2]
         return @@level.to_s
      end
   end
end # Class Drop
# -------------------------------------------------------------------------------------------------
####  CONTEXT MENUS  ####
# DROP.RB  (Drop Group/Component On Selected Level (Change In Z))  
#    ab.add_separator if (Drop.validate_level_entity || Drop.validate_drop_entity)
#     rm.add_item("Drop at intersection") { Drop.drop_intersection } if Drop.validate_drop_entity
#     rm.add_item("Drop at ...") { Drop.drop_at } if Drop.validate_drop_entity
#     rm.add_item("Get position: "+Drop.get_plane_position) {} if Drop.validate_level_entity
#end
#***********************************************************************************************************************************************************************************
#                       WELD
#***********************************************************************************************************************************************************************************
# Copyright 2004,2005 by Rick Wilson 
# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.
# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Name :          Weld 2.3
# Description :   weld edges into curve ("polyline")
# Author :        RickW
# Usage :         select edges, then run the script (Plugins>Weld)
#			Edges will be joined into a curve ("polyline")
#			Script will prompt to find faces
# Date :          4.Aug.2004
# Type :		tool
# History:
#			3.0 (4.Aug.2005) -  Special Anniversary Edition! :)
#						  now welds to existing curves, adds option
#						  to find faces
#			2.2 (1.Aug.2005) -  used group/explode to fix most problems
#						  still won't weld to non-arc curves
#			2.1 (27.Jul.2005) - fixed the fix that caused strange non-welding
#						  issues.
#			2.0 (25.Jul.2005) - fixed bug where unselected edges were welded,
#						  now retains existing faces.
#			1.2 (12.May.2005) - fixed edge bug that occasionally resulted
#						  in strange edge locations and deletions
#			1.1 (10.Sep.2004) - allows working in groups/components
#			1.0 (4.Aug.2004) - first version
def join_edges
	model=Sketchup.active_model
	ents=model.active_entities
	sel=model.selection
	sl=sel.length
	verts=[]
	edges=[]
	newVerts=[]
	startEdge=startVert=nil
#CHECK FOR A VALID SELECTION
#	if sel.first.curve
#		return UI.messagebox("Please select at least 2 non-joined edges") if sel.length == sel.first.curve.count_edges
#	end
#DELETE NON-EDGES, GET THE VERTICES
	sel.each {|item| edges.push(item) if item.typename=="Edge"}
	edges.each {|edge| verts.push(edge.vertices)}
	verts.flatten!
	return UI.messagebox("Please select at least 2 edges") if edges.length < 2
#FIND AN END VERTEX
	vertsShort=[]
	vertsLong=[]
	verts.each do |v|
		if vertsLong.include?(v)
			vertsShort.push(v)
		else
			vertsLong.push(v)
		end
	end
	if (startVert=(vertsLong-vertsShort).first)==nil
		startVert=vertsLong.first
		closed=true
		startEdge = startVert.edges.first
	else
		closed=false
		startEdge = (edges & startVert.edges).first
	end
	#sel.clear
#SORT VERTICES, LIMITING TO THOSE IN THE SELECTION SET
	if startVert==startEdge.start
		newVerts=[startVert]
		counter=0
		while newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==newVerts.last
					newVerts.push(edge.start)
				elsif edge.start==newVerts.last
					newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				newVerts.reverse!
				reversed=true
			end
		end
	else
		newVerts=[startVert]
		counter=0
		while newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==newVerts.last
					newVerts.push(edge.start)
				elsif edge.start==newVerts.last
					newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				newVerts.reverse!
				reversed=true
			end
		end
	end
	newVerts.uniq!
	newVerts.reverse! if reversed
#CONVERT VERTICES TO POINT3Ds
#	newVerts.collect!{|x| x.position}
#	if closed
#		newVerts.push(newVerts[0])
#	else
#		newVerts.push(newVerts[0]) if UI.messagebox("Close curve?", MB_YESNO)==6
#		closed=true
#	end
#CREATE THE CURVE
	model.start_operation "weld"
	edges.each {|item| item.explode_curve if item.curve and newVerts.include?(item.start.position)}
#sel.each {|e| e.erase!}
	gp1e=(gp1=ents.add_group).entities
	curveEdges=gp1e.add_curve(newVerts)
	gp1.explode
	ents[-1].find_faces #if closed and ents[-1].typename=="Edge" and UI.messagebox("Find faces for this curve?", MB_YESNO, "Weld Edges Utility")==6
	model.commit_operation
	sel.clear
end
#***********************************************************************************************************************************************************************************
#                       SKETCHY BEVEL
#***********************************************************************************************************************************************************************************
#SketchyBevel 
#Bevel all selected faces.
#Adds a context menu item called Bevel
#
#
#Copyright 2008 Chris Phillips
module SketchyBevel
    if($bevel_menu_loaded==nil)
        $bevel_menu_loaded=true
        
        if(Sketchup.active_model.options["UnitsOptions"]["LengthUnit"]<2)
            $bevelSettings = [1.0.inch, "true"]
        else
            $bevelSettings = [5.0.cm, "true"]
        end           
        UI.add_context_menu_handler { |menu|
            selection=Sketchup.active_model.selection
            
            if(!selection.empty?)
		
            sbb = menu.add_submenu("- Sketchy Bevel")
                
                sbb.add_item("Bevel"){            
                    prompts = ["Amount","Cap Holes"]#,"Interations"]
                    results = inputbox(prompts, $bevelSettings,["","true|false"], "Bevel selection")
                    if (results)
                        $bevelSettings=results
                        Sketchup.active_model.selection.bevel(results[0],results[1])
                    end
                }
                sbb.add_item("Cap Holes"){
                    Sketchup.active_model.start_operation("Cap Holes")                
                    Sketchup.active_model.selection.cap_holes()
                    Sketchup.active_model.commit_operation()
                }
            end
        }
    end
    class Sketchup::Selection
        
        def find_adjcent_coplaner_edges(edge,plane)
            edges=[]
            edge.vertices.each{|v|
                v.edges.each{|e|
                    next if(e==edge)
                    if(!e.get_attribute("SPCAP","used",false)&&
                            e.start.position.on_plane?(plane)&& e.end.position.on_plane?(plane))
                        edges.push(e)
                        e.set_attribute("SPCAP","used",true)
                        edges.push(find_adjcent_coplaner_edges(e,plane))
                    end
                }
            }
            return edges.flatten
                    
        end
        
        def cap_holes()
            self.faces.each{|f|
                    f.edges.each{|e|e.delete_attribute("SPCAP","used")}
                }
            
            newfaces=[]
            while(true)
                caps=[]
                self.faces.each{|f|
                    f.edges.each{|e|
                        if(e.faces.length<2)
                            e.vertices.each{|v|
                                v.edges.each{|ve|
                                    if(e!=ve && ve.faces.length<2 && 
                                            !e.get_attribute("SPCAP","used",false) && !ve.get_attribute("SPCAP","used",false) )
                                            
                                            
                                        #make face out of e and ve
                                        #e.set_attribute("SPCAP","used",true)
                                        #ve.set_attribute("SPCAP","used",true)
                                        pts=[v.position,e.other_vertex(v).position,ve.other_vertex(v).position]
                                        plane=Geom.fit_plane_to_points(pts)
                                        ce=find_adjcent_coplaner_edges(ve,plane)
                                        #ce.push(ve)
                                        #puts ce.inspect
                                        caps.push(ce)
                                    end
                                }
                            }
                        end
                    }
                }
                caps.each{|f|
                    if(f.length==2)
                        nf= [f[0].start.position.to_a,f[0].end.position.to_a,f[1].start.position.to_a,f[1].end.position.to_a].uniq
                        #puts "tri"
                        #puts nf.inspect
                        newfaces.push(Sketchup.active_model.active_entities.add_face(nf))
                    else
                        #puts "quad"
                        #puts f.inspect
#puts f                        
#cf=f.select{|tf|!tf.deleted?}
                        newfaces.push(Sketchup.active_model.active_entities.add_face(f))
                    end
                }
                #self.clear()
                #self.add(caps)
                #puts newfaces
                #self.add(newfaces) if(!newfaces.empty?)
                self.faces.each{|f|
                    f.edges.each{|e|e.delete_attribute("SPCAP","used")}
                }
                puts caps.length
                break# if(caps.length==0)#found no caps so exit loop.
                    
            end                
            return newfaces
        end
        def old_cap_holes()
            newfaces=[]
            while(true)
                caps=[]
                self.faces.each{|f|
                    f.edges.each{|e|
                        if(e.faces.length<2)
                            e.vertices.each{|v|
                                v.edges.each{|ve|
                                    if(e!=ve && ve.faces.length<2 && 
                                            !e.get_attribute("SPCAP","used",false) && !ve.get_attribute("SPCAP","used",false) )
                                        #make face out of e and ve
                                        e.set_attribute("SPCAP","used",true)
                                        ve.set_attribute("SPCAP","used",true)
                                        caps.push([v.position,e.other_vertex(v).position,ve.other_vertex(v).position])
                                    end
                                }
                            }
                        end
                    }
                }
                caps.each{|f|newfaces.push(Sketchup.active_model.active_entities.add_face(f))}
                self.add(newfaces)
                self.faces.each{|f|
                    f.edges.each{|e|e.delete_attribute("SPCAP","used")}
                }
                puts caps.length
                break if(caps.length==0)#found no caps so exit loop.
                    
            end                
            return newfaces
        end
        
        
        def bevel(amount,bCapHoles)
            Sketchup.active_model.start_operation("Bevel")
            
            #build a nested array of faces[edges[verts]]
            allFaces=self.faces.collect{|f|
                verts=f.outer_loop.vertices
                edges=[]
                i=0
                while i<verts.length
                    edges.push([verts[i].position,verts[(i+1)%verts.length].position])
                    i=i+1
                end
                edges
            }
            #make a simple array of all the points        
            allPoints=allFaces.flatten
            #find any duplicate points. These points will make cap faces later.
            dupePoints=allPoints.collect{|p|allPoints.select{|p2|p==p2}}
            
            #simple array of all edges
            allEdges=[]
            allFaces.each{|f|
                f.each{|edge|allEdges.push(edge)}
            }
            
            #find overlapping edges. Each will be a bevel face
            dupeEdges=[]
            allEdges.each_with_index{|e,i|
                allEdges.each_with_index{|e2,i2| 
                    if(i!=i2&& (e[0]==e2[0]&&e[1]==e2[1])||(e[1]==e2[0]&&e[0]==e2[1]))
                        dupeEdges.push([e,e2])
                    end
                }
            }
            
            shrunkFaces=[]
            #Create a shrunken version of each face
            allFaces.each{|f|
                SketchyBevel::shrink_face(f,amount)
                pts=f.collect{|edge|edge[0]}
                shrunkFaces.push(Sketchup.active_model.active_entities.add_face(pts))
            }
            #Create the bevel. A 4 sided polygon made from each formerly overlapping edge
            newFaces=[]
            dupeEdges.each{|edge|
                newFaces.push(Sketchup.active_model.active_entities.add_face(edge.flatten))
            }
            
bCapHoles="false"
            #~ caps=[]
            #~ capEdges=newFaces.each{|f|
                #~ f.edges.each{|e|
                    #~ if(e.faces.length<2)
                        #~ e.vertices.each{|v|
                            #~ v.edges.each{|ve|
                                #~ if(!e.get_attribute("SPBEVEL","alreadybeveled",false) && 
                                    #~ !ve.get_attribute("SPBEVEL","alreadybeveled",false) && e!=ve && ve.faces.length<2)
                                    #~ #make face out of e and ve
                                    #~ e.set_attribute("SPBEVEL","alreadybeveled",true)
                                    #~ ve.set_attribute("SPBEVEL","alreadybeveled",true)
                                    #~ caps.push([v.position,e.other_vertex(v).position,ve.other_vertex(v).position])
                                #~ end
                            #~ }
                        #~ }
                    #~ end
                    #~ }
                #~ }
                
            #~ caps.each{|f|Sketchup.active_model.entities.add_face(f)}
            #capEdges=newFaces.collect{|f|
            #    f.edges.select{|e|e.faces.length<2}
            #    }
            #puts "caps"
            #puts capEdges.inspect
            
            #~ #cap the corners by drawing a face around each formerly duplicated point.
            dupePoints.each{|pts|
                #remove any duplicated points.
                tpts=pts.collect{|p|p.to_a}.uniq
                case tpts.length
                when 3
                    Sketchup.active_model.active_entities.add_face(tpts)
                when 4
                    0.upto(tpts.length-2){|i|
                        tri=[tpts[i],tpts[(i+1)%tpts.length],tpts[(i+2)%tpts.length]]
                        Sketchup.active_model.active_entities.add_face(tri)    
                    }   
                
                    #Sketchup.active_model.entities.add_face(tpts[0],tpts[1],tpts[2])
                    #Sketchup.active_model.entities.add_face(tpts[1],tpts[2],tpts[3])
                    #ca=(tpts[2]-tpts[1]).cross(tpts[3]-tpts[2])
                    #if((tpts[3]-tpts[2]).cross(tpts[0]-tpts[3]).dot(ca)<0.0)
                        #Sketchup.active_model.entities.add_face(tpts[2],tpts[3],tpts[0])
                    #else
                        #Sketchup.active_model.entities.add_face(tpts[1],tpts[3],tpts[0])
                    #end
                    #Sketchup.active_model.entities.add_face(tpts.slice(0,3))
                end
            } if(bCapHoles=="true")
            #remove all the old geometry.
            Sketchup.active_model.selection.each{|e|e.erase!}
            
            Sketchup.active_model.selection.add(shrunkFaces)            
            Sketchup.active_model.selection.add(newFaces)
            Sketchup.active_model.selection.add(cap_holes())
            Sketchup.active_model.commit_operation()
        end
        def faces()
            self.select{|s|s.class==Sketchup::Face}    
        end
    end
    
    
    def self.shrink_face(face,amount)
        
        #TODO.face is really edges[].
        
        
        normal=(face.first[1]-face.first[0]).cross(face.last[1]-face.last[0]).normalize
        if(normal.length==0)
            plane=Geom.fit_plane_to_points([face[0][0],face[0][1],face[1][0],face[1][1],face[2][0],face[2][1]])
            puts normal
            normal=Geom::Vector3d.new(plane[0],plane[1],plane[2])
            puts normal
        end
        #normal=face.normal
        #Sketchup.active_model.entities.add_line(face.first[0],face.first[0]+normal)
        
        #find the direction to move each vert in face
        shrinkVectors=[]
        lastEdge=face.last #use last vert in face is connected to the first vert in face.
        face.each{|edge|
            #create 2 vectors from each edge
            va=lastEdge[0]-lastEdge[1]
            vb=edge[1]-edge[0]
            lastEdge=edge #use this next loop.
            #find angle between the two vectors.        
            if(va.cross(vb).dot(normal)<0) #check to see if angle >180
                angle=(360.degrees-va.angle_between(vb)) #angle is >180
            else
                angle=va.angle_between(vb)
            end
            #rotate one vector 1/2 the angle between the two edges
            sv=va.transform(Geom::Transformation.new([0,0,0],normal,angle/2))
            #set the distance to move based on the amount of bevel.
            sv.length=amount/Math.sin(angle/2)
            #save.
            shrinkVectors.push(sv)
        }
        
        edge=face[0]
        before=(edge[0]-edge[1]).length
        after=(edge[0].transform(shrinkVectors[0])-edge[1].transform(shrinkVectors[1])).length
        
        
        bFlipped=false
        #~ if(edge[0]!=edge[1]&& (edge[0].transform(shrinkVectors[0])-edge[1].transform(shrinkVectors[1])).length>before)
            #~ bFlipped=true
        #~ end
        
        #now move each edge by the amount calculated. This actually shrinks the face.
        face.each_index{|index|
            edge=face[index]
            if(!bFlipped)
                edge[0].transform!(shrinkVectors[index])
                edge[1].transform!(shrinkVectors[(index+1)%face.length])
            else
                edge[0].transform!(shrinkVectors[index].reverse)
                edge[1].transform!(shrinkVectors[(index+1)%face.length].reverse)
            end
        }
    end
end
#***********************************************************************************************************************************************************************************
#                       CHAMFER ALONG PATH
#***********************************************************************************************************************************************************************************
###############################################################################
#       SCRIPT NAME : Chamfer_Along_Path.rb
#
#       AUTHOR : Ashley Joyce (c) 2006  Version: 1.0
#
#       LICENSING: Smustard.com Ruby Script End User License Agreement
#
#       Based on ExtrudeAlongPath.rb written by Tig which itself was based on
#		original wall making idea by Didier Bur and using some vertex array
#		ideas from Rick Wilson's Weld.rb
#
#       USAGE :
#		Select joined lines, arcs, circles, curves, etc.
#		Select "Chamfer Along Path" from the Tool menu.
#
#       This tool differs from ExtrudeAlong Path in that the extrude face is
#       a triangular shape with a specified width and height. The Triangular
#       face is referenced from the edge. Branching paths will not
#       extrude.
#
#       The only tricky bit about this tool is knowing that (+) positive and
#       (-) negative values for both width and height parameters change the
#       orientation of the triangular face.
#
#       This means you can create (4) different orientations by changing
#       the Height and Width Settings using (+) positive & (-) negative values
#
#       TUTORIAL & UPGRADES:
#
#       FOR GRAPHICAL TUTORIALS ON USING THE CHAMFER ALONG PATH TOOL AND FUTURE
#       UPGRADES PLEASE VISIT  WWW.SMUSTARD.COM
#
#       Email: ashley@bowin.com.au
#
#
#***********************************************************************************************************************************************************************************
#  CHAMFER ALONG PATH
#***********************************************************************************************************************************************************************************
def chamfer_along_path
@error=0
model = Sketchup.active_model
entities = model.active_entities
ss = model.selection
### show VCB and status info
Sketchup::set_status_text("Chamfer Along Path...ENTER PARAMETERS...", SB_PROMPT)
Sketchup::set_status_text(" ", SB_VCB_LABEL)
Sketchup::set_status_text(" ", SB_VCB_VALUE)
###
if ss.empty?
  Sketchup::set_status_text("Chamfer Along Path... NO SELECTION ! ", SB_PROMPT)
  UI.messagebox("No Selection to Chamfer.")
  return nil
end
###############################################################
def get_vertices
###############################################################
### this next bit is mainly thanks to Rick Wilson's weld.rb
### & Tig's ExtrudeAlongPath.rb
###############################################################
	model=Sketchup.active_model
	ents=model.active_entities
	@sel=model.selection
	sl=@sel.length
	verts=[]
	edges=[]
	@newVerts=[]
	startEdge=startVert=nil
#DELETE NON-EDGES, GET THE VERTICES
	@sel.each {|item| edges.push(item) if item.typename=="Edge"}
	edges.each {|edge| verts.push(edge.vertices)}
	verts.flatten!
#FIND AN END VERTEX
	vertsShort=[]
	vertsLong=[]
	vertsEnds=[] ### 17/8/5 ### to ensure array is only ends ###
	verts.each do |v|
		if vertsLong.include?(v)
			vertsShort.push(v)
		else
			vertsLong.push(v)
		end
	end
	vertsLong.each do |v| ### 17/8/5 ### to ensure array is only ends ###
		if not vertsShort.include?(v)
			vertsEnds.push(v)
		end
	end ### 17/8/5 ### to ensure array is only ends ###
	if vertsEnds.length==0 ### i.e. it's looped ###
 		### path start or end ? ###
		if @theEnd==0
		  startVert=vertsLong.first
		  startEdge=startVert.edges.first
		else
		  startVert=vertsLong.last
		  startEdge=startVert.edges.first
		end
		###
		closed=true
	else
		if vertsEnds.length != 2
			Sketchup::set_status_text("Chamfer Along Path... PATH ERROR ! ", SB_PROMPT)
			UI.messagebox("The selected Path either branches or is not continuous ! \nCannot make a Chamfer ! \nRe-select a single continuous path... ") ### 17/8/5 ###
			@error=1
			return nil
		else
			### path start or end ?
			if @theEnd==0
			  startVert=vertsEnds.first
			else
			  startVert=vertsEnds.last
			end
			###
			closed=false
			startEdge=startVert.edges.first
		end
	end
	@sel.clear
#SORT VERTICES, LIMITING TO THOSE IN THE SELECTION SET
	if startVert==startEdge.start
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Chamfer Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	else
		@newVerts=[startVert]
		counter=0
		while @newVerts.length < verts.length
			edges.each do |edge|
				if edge.end==@newVerts.last
					@newVerts.push(edge.start)
				elsif edge.start==@newVerts.last
					@newVerts.push(edge.end)
				end
			end
			counter+=1
			if counter > verts.length
				Sketchup::set_status_text("Chamfer Along Path... ERROR ! ", SB_PROMPT)
				return nil if UI.messagebox("There seems to be a problem. Try again?", MB_YESNO)!=6
				@newVerts.reverse!
				reversed=true
			end
		end
	end
	@newVerts.reverse! if reversed
#CONVERT VERTICES TO POINT3Ds
	@newVerts.collect!{|x| x.position}
### now have an array of vertices in order with NO forced closed loop ...
end ### get_vertices
@theEnd=0
get_vertices
if @error==1
  return nil
end
### Input Parameters ################################################
#####################################################################
prompts = ["Width: ","Height: "]
if not $widthIn
   values = [-10.mm,-10.mm]
else
   values = [$widthIn,$heightIn]
end
results = inputbox prompts, values,"Chamfer Parameters (mm)"
return nil if not results ### i.e. the user cancelled the operation
$widthIn,$heightIn = results
###################################################################
### this next bit is mainly thanks to Tig's ExtrudeAlongPath.rb ###
###################################################################
pt1 = @newVerts[0]
pt2 = @newVerts[1]
width = $widthIn
if $widthIn == 0.mm ### can't be 0 ###
   $widthIn = -10.mm
   UI.messagebox("Zero Width NOT allowed ! ")
   return nil
end
###
if $heightIn == 0.mm ### can't be 0 ###
   $heightIn = -10.mm
   UI.messagebox("Zero Height NOT allowed ! ")
   return nil
end
###
if (pt1.x == pt2.x) and (pt1.y == pt2.y) ### vertical 1st path ###
   vflag = 1
   height = $heightIn
else
   vflag = 0
   height = (($heightIn * (pt1.distance pt2)) / (pt1.distance [pt2.x,pt2.y,pt1.z]))
end
   offL = width
   offR = 0
   heightUp = height
   heightDn = 0
if vflag == 1 ###
  vec = pt1.vector_to [(pt1.x + 1), pt1.y, pt1.z]
else
  vec = pt1.vector_to [pt2.x, pt2.y, pt1.z]
end
piBy2 = 1.5707963267948965 ### a radian right-angle for calculating vector offset to edges
rotated_vecL = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 + piBy2)))
pt1_leftC = pt1.offset(rotated_vecL, offL)
rotated_vecR = vec.transform(Geom::Transformation.rotation(pt1, [0,0,1], (0 - piBy2)))
pt1_rightC = pt1.offset(rotated_vecR, offR)
if vflag == 1 ### 1st path is vertical
  pt1_left  = pt1_leftC.offset([-1,0,0], heightDn)
  pt1_right = pt1_rightC.offset([-1,0,0], heightDn)
  pt2_left  = pt1_leftC.offset([1,0,0], heightUp)
  pt2_right = pt1_rightC.offset([1,0,0], heightUp)
else
  pt1_left  = pt1_leftC.offset([0,0,-1], heightDn)
  pt1_right = pt1_rightC.offset([0,0,-1], heightDn)
  pt2_left  = pt1_leftC.offset([0,0,1], heightUp)
  pt2_right = pt1_rightC.offset([0,0,1], heightUp)
end
ents=entities
##############################################
model.start_operation "MakeChamberface"
##############################################
  theChamberFace = ents.add_face(pt1_left,pt1_right,pt2_right)
  theChamberFace.reverse!
  theChamberEdges= []
  0.upto((@newVerts.length/2 -1)) do |i|
   theChamberEdges[i] = ents.add_line(@newVerts[i],@newVerts[i+1]) ### make vertices into Chamber edges
 end
#############################################
model.commit_operation
#############################################
  begin
  theExtrusion=theChamberFace.followme theChamberEdges
  rescue
  UI.messagebox $!.message
  end
end ### end def chamfer_along_path
###
### #################################################################
#***********************************************************************************************************************************************************************************
#                       PROJECTIONS
#***********************************************************************************************************************************************************************************
#        Name : projection.rb
# Description : 
#      Author : Didier Bur
#       Usage : 
#        Date : 11.2005
#        Type : geometric tools set
#              : 10.11.07 updated
# more user-friendly commands (click points instead of selecting objects for plane projections and vectors
# project faces along Z and along vector gathered in one command.
# ExtrudelineTool included
# lines2cyl and credits suppressed
#Cline at intersection of faces enhanced
# new icons
#              : 01.08.08 updated
# modified CpointTool, added normal line from cpoint
def create_faces_dialog
$proj_create_faces = "No" if not $proj_create_faces
create_faces = ["Yes","No"]
enums = [create_faces.join("|")]
prompts = ["Create face(s): "]
values = [$proj_create_faces]
results = inputbox prompts, values, enums, "Projection Options"
return if not results
$proj_create_faces = results[0]
end
def create_lines_dialog
$proj_create_lines = "No" if not $proj_create_lines
create_lines = ["Yes","No"]
enums = [create_lines.join("|")]
prompts = ["Create line(s): "]
values = [$proj_create_lines]
results = inputbox prompts, values, enums, "Projection Options"
return if not results
$proj_create_lines = results[0]
end
# Removes from selection all objects of types contained in object_types array
def sel_remove(sel, object_types)
  sel.each do |obj|
    if object_types.include? obj.typename
      sel.remove(obj)
    end
  end
return sel
end
# Removes from selection all objects of types NOT contained in object_types array
def sel_keep(sel, object_types)
  sel.each do |obj|
    if (not object_types.include? obj.typename)
      sel.remove(obj)
    end
  end
return sel
end
#***********************************************************************************************************************************************************************************
#                       PUSHPULL MULTIPLE FACES
#***********************************************************************************************************************************************************************************
# --------------------------------------------------- Extrudes a (set of) face(s) along their normals
def extrude_faces_along_normal
model=Sketchup.active_model
model.start_operation "Push-Pull along normal"
ss=model.selection
$pp_dist = 1 if not $pp_dist
prompts = ["Push-Pull distance: "]
value = [$pp_dist]
result = inputbox prompts, value, "Push-Pull options"
return if not result
$pp_dist = result[0]
ss=sel_keep(ss,["Face"])
#Sketchup.active_model.selection.each{|e| if e.typename == "Face" then
ss.each do |e|
  #e.pushpull Float( dist[0])
  e.pushpull($pp_dist.to_l)
end
model.commit_operation
end
#***********************************************************************************************************************************************************************************
#                       PUSHPULL ALONG VECTOR
#***********************************************************************************************************************************************************************************
class PushPullFacesVectorTool
def initialize
    $points=[]
    @ip1 = nil
    @ip2 = nil
    @xdown = 0
    @ydown = 0
    @state = 0
    @model = Sketchup.active_model
    @model.start_operation "Push-Pull faces along vector"
    @entities = @model.active_entities
    @ss = @model.selection
end
def activate
    Sketchup.active_model.start_operation "Push-Pull faces along vector"
    @ip1 = Sketchup::InputPoint.new
    @ip2 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false
    self.reset(nil)
    
    case @ss.length
      when 0
        UI.messagebox("No selection.\nPlease select faces.")
        Sketchup.active_model.commit_operation
        Sketchup.send_action "selectSelectionTool:"
        return
      when 1
        if @ss[0].typename != "Face"
          UI.messagebox("You have selected only one " + @ss[0].typename + ".\nPlease select faces.")
          @ss.clear
          Sketchup.active_model.commit_operation
          Sketchup.send_action "selectSelectionTool:"
        end
        return  
    end
end
def deactivate(view)
    view.invalidate if @drawn
end
def onMouseMove(flags, x, y, view)
@ip.pick view, x, y
  if( @ip != @ip1 )
    view.invalidate if( @ip.display? or @ip1.display? )
    @ip1.copy! @ip
  end
        
case @state
    when 0
      view.tooltip = "First point of vector"
      Sketchup::set_status_text("First point of vector", SB_PROMPT)
    when 1
      view.tooltip = "Second point of vector"
      Sketchup::set_status_text("Second point of vector", SB_PROMPT)
    when 2
      view.tooltip = ""
      Sketchup::set_status_text("", SB_PROMPT)
  end
end
def onLButtonDown(flags, x, y, view)
  case @state
    when 0
      @xdown = x
      @ydown = y
      @ip1.pick view, x, y
      view.tooltip = @ip1.tooltip if( @ip1.valid? )
      $points.push @ip1.position
      @state = 1
      view.invalidate
    when 1
      @xdown = x
      @ydown = y
      @ip2.pick view, x, y, @ip1
      view.tooltip = @ip2.tooltip if( @ip2.valid? )
      $points.push @ip2.position
      @state = 2
      view.invalidate 
    when 2
      # Clear any inference lock
      view.lock_inference
  end #case
end
# The onLButtonUp method is called when the user releases the left mouse button.
def onLButtonUp(flags, x, y, view)
self.reset(view)
self.create_geometry(view) if @state == 2
end
def onKeyDown(key, repeat, flags, view)
if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
        # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
            view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
            view.lock_inference @ip2
        end
    end
end
def onKeyUp(key, repeat, flags, view)
return if not @state == 3
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end
def onUserText(text, view)
    return if not @state == 1
    #return if not @ip2.valid?
    
    begin
        @value = text.to_l
    rescue
        # Error parsing the text
        UI.messagebox("Cannot convert #{text} to a length.")
        @value = nil
        Sketchup::set_status_text "", SB_VCB_VALUE
    end
    return if !@value
    @pt1 = @ip1.position
    @vec = @ip2.position - @pt1
    if( @vec.length == 0.0 )
        return
    end
    @vec.length = @value
    @pt2=@pt1.offset @vec
    #puts @vec
    #puts @vec.length.to_s
    #puts @pt1
    #puts @pt2
    
    $points.push(@pt2)
    self.create_geometry(view)
    self.reset(view)
end
def draw(view)
    if( @ip1.valid? )
        if( @ip1.display? )
            @ip1.draw(view)
            @drawn = true
        end
    end
    
  if $points.length == 1
    view.draw_points $points[0], 8, 1, "black"
  end
  
  if $points.length == 2
    view.line_width = 2
    view.draw_points $points[0], 8, 1, "black"
    view.draw_points $points[1], 8, 1, "black"
  end
  view.lock_inference
end
# onCancel is called when the user hits the escape key
def onCancel(flag, view)
self.reset(view)
    self.draw_geometry(view) if @state < 2
    if $points.length < 2
      # We got less than 5 points
      UI.messagebox($points.length.to_s + " points are not enough for this projection !")
      self.reset(view)
    end
end
# Reset the tool back to its initial state
def reset(view)
    #@state = 0
    Sketchup::set_status_text("First point of vector", SB_PROMPT)
    @ip1.clear
    @ip2.clear
    if( view )
        view.tooltip = nil
        view.invalidate if @drawn
    end
    @drawn = false
    @dragging = false
end
# Create new geometry when the user has picked five points.
def create_geometry(view)
#model = Sketchup.active_model
#model.start_operation "Push-Pull faces along vector"
#entities = model.active_entities
ss = sel_keep(@model.selection, ["Face"])
vect = Geom::Vector3d.new($points[1] - $points[0])
t = Geom::Transformation.translation(vect)
# Push-Pull, process each face
current_face = 0
pt_vector_1 = $points[0]
pt_vector_2 = $points[1]
delta_x = pt_vector_2[0] - pt_vector_1[0]
delta_y = pt_vector_2[1] - pt_vector_1[1]
delta_z = pt_vector_2[2] - pt_vector_1[2]
ss.each do |face| 
  pts = face.vertices
  length_pts = pts.length
  # copy the selected face
  pts2 = []
  pts.each do |pt|
    pts2.push(pt.position.transform(t))
  end
  face_copy = @entities.add_face pts2
  pts_copy = face_copy.vertices
  # create faces between pairs of vertices of each face
  i = 0
  j = 0
  0.upto(pts.length - 2) do |new_face| 
    new_face = @entities.add_face(pts[i], pts_copy[i], pts_copy[i + 1], pts[i + 1])
    i = i + 1
  end    # of upto
  # close the shape
  new_face = @entities.add_face(pts[pts.length - 1], pts_copy[pts.length - 1], pts_copy[0], pts[0])
end
@model.commit_operation
Sketchup.send_action "selectSelectionTool:"
end
# Draw the geometry (temp)
def draw_geometry(pt1, pt2, view)
    view.draw_line(pt1, pt2)
end
end # class PushPullFacesVectorTool
#***********************************************************************************************************************************************************************************
#                       RANDOM PUSHPULL
#***********************************************************************************************************************************************************************************
# --------------------------------------------------- Random pushpull within interval
def randomPushPull
	model=Sketchup.active_model
	model.start_operation "Random Push-Pull within interval"
	ss=model.selection
	if ss.empty? 
  		UI.messagebox("No Selection.\nPlease select faces.")
  		return nil
	end
  ss = sel_keep(ss,["Face"])
	if not $ppmax
		$ppmax=2.0 ### numbers in current units currently assumes mm - see below***
		$ppmin=1.0
	end
	prompts = ["Height (max) :  ", "Height (min) :  "]
	value =   [$ppmax, $ppmin]
	results = inputbox prompts, value, "Random Push-Pull Range (mm)" ###***
	return if not results
	$ppmax=results[0]
	$ppmin=results[1]
	ss.each do |obj|
		if obj.typename == "Face"
			ht=(($ppmax - $ppmin) * rand)
			obj.pushpull -ht #.mm ### into mm ***
		end
	end
	model.commit_operation
end
#***********************************************************************************************************************************************************************************
#                       COMPONENTS SPRAY TOOL
#***********************************************************************************************************************************************************************************
require 'sketchup.rb'

class ProgressBar 

@@err_total_notnumeric = "ProgressBar: Total must be a positive integer."
@@err_count_notnumeric = "ProgressBar: Iteration Count must be numeric"
@@end_time             = "End @ " 
@@progresschar = ">" ; 
@@initial_block = "-" * 50     # Default progress bar line sequence.

def initialize(total,phase=nil)  
  if (!total.integer? or total < 0)
    raise(ArgumentError,@@err_total_notnumeric) 
    return ; 
    end ; 
  @total = total.to_i ; 
  @phase = phase ; 
  @firsttime = true ; 
  end ; 

def update(iteration) 
  if !iteration.integer? 
    raise(ArgumentError,@@err_count_notnumeric)  
    return ; 
    end ; 
  iteration = [iteration.abs,@total].min  # make sure we don't exceed the total count or have a value < 0
  pct = [1,(iteration*100)/@total].max    # Calculate percentage complete.
                                          # round up to 1% if anything less than 1%.
  end_time = "?" ; 
  if @firsttime then ; 
    # Get the current time of day. 
    # set up an elapsed timer so we can calculate expected end time.  
    @time1 = Time.now ;     # Get current time of day. 
    @firsttime = false ;    # turn off switch 
  else 
    # divide the elapsed time by the pct complete, then multiple that by 100, then add that to the 
    # start time, and that is the expected end time.  
    end_time = Time.at(((((Time.now-@time1).to_f)/pct)*100 + @time1.to_f).to_i).strftime("%H:%M:%S") 
    end ; 
  pct_pos = [pct/2,1].max ;
  current_block = @@initial_block[0,pct_pos-1] << @@progresschar << @@initial_block[pct_pos,@@initial_block.length]
  Sketchup.set_status_text(current_block << "   " << (pct.to_s) << "%. #{@@end_time} " << end_time << " #{@phase}")
  #Sketchup.set_status_text(current_block << "   " << (pct.to_s) << "%")
  if (((Time.now-@time1).to_f/pct)*100) > $cbt_delay
    $cbt_stop_timer = true
  end
  #if (((Time.now-@time1).to_f > 300
    #$cbt_stop_timer = true
  #end
  end ; 

end ; # class ProgressBar ;

#----------------------------------------------------------------------------- Returns highest/lowest Z of model
def cbt_model_top
model = Sketchup.active_model
return model.bounds.max.z
end
def cbt_model_bottom
model = Sketchup.active_model
return model.bounds.min.z
end
#----------------------------------------------------------------------------- Translate an array of points to altitude h
def map_z(points_array, h)
mapped_array =[]
points_array.each { |p| mapped_array.push Geom::Point3d.new(p.x, p.y, h) }
return mapped_array
end
#----------------------------------------------------------------------------- Returns average width and depth of component definition
def cbt_average_bounds(comps_def_array)
  area = 0.0
  average_scale = ($cbt_low_scale.abs + $cbt_high_scale.abs) / 2.0
  comps_def_array.each { |cd| area = area + cbt_area_bounds(cd) }
  return (area * average_scale) / comps_def_array.length
end
#----------------------------------------------------------------------------- Returns average width and depth of component definition
def cbt_average_widths(comps_def_array)
  w = 0.0
  average_scale = ($cbt_low_scale.abs + $cbt_high_scale.abs) / 2.0
  comps_def_array.each do |cd|
    w += cd.bounds.width if cd.bounds.width > 0.001
    w += [cd.bounds.width,cd.bounds.depth,cd.bounds.height].max if cd.bounds.width < 0.001
  end
  return (w * average_scale) / comps_def_array.length if w!=0
end
#----------------------------------------------------------------------------- Returns area of component definition
def cbt_area_bounds(c)
  w=c.bounds.width
  d=c.bounds.depth
  if w<0.001 and d>0.001
    w=d/10
  end
  if w>0.001 and d<0.001
    d=w/10
  end
  if w>0.001 and d>0.001
    return w*d
  else
    return c.bounds.height * c.bounds.height
  end
  #return (c.bounds.width * c.bounds.depth * c.bounds.height) * 0.6666
end
#----------------------------------------------------------------------------- Returns BB of faces set
def cbt_faces_bounds(faces)
Sketchup.set_status_text("Computing faces bounds...")
fbb = Geom::BoundingBox.new
faces.each { |f| fbb.add f.bounds }
return fbb
end
#----------------------------------------------------------------------------- Returns area of faces set
def cbt_faces_area(faces)
Sketchup.set_status_text("Computing faces area...")
a = 0.0
faces.each { |f| a = a + f.area }
return a
end
#----------------------------------------------------------------------------- Returns slope angle of face [0->PI/2]
def cbt_slope(hit_objet)
case hit_objet[1][0].typename
  when "Face"
    f = hit_objet[1][0]
  when "ComponentInstance"
    f = hit_objet[1][1]
  when "Group"
    f = hit_objet[1][1]
  when "Image"
    f = hit_objet[1][1]
  when "Edge"
    p = hit_objet[1][0].start.position
    pz = hit_objet[1][0].end.position
    p3 = Geom::Point3d.new(pz.x,pz.y,p.z)
    a = hit_objet[1][0].line[1].angle_between p.vector_to(p3)
    return a
  else
    return nil
end
if f.typename == "Face"
  n = f.normal
  p = Geom::Point3d.new(0,0,0)
  pz = p.offset(n)
  p0 = Geom::Point3d.new(pz.x,pz.y,0)
  a = p.vector_to(p0).angle_between n
  return a
else
  return 0.0
end
end
#-----------------------------------------------------------------------------  Test if low_steep < slope of face < high_steep
def cbt_match_slope(f)
return true if $cbt_low_steep == 0 and $cbt_high_steep == 90
l = Math::sin($cbt_low_steep.degrees)
h = Math::sin($cbt_high_steep.degrees)
fa = Math::cos(cbt_slope(f))
if l <= fa and h > fa
  return true
  else
  return false
end
end
#-----------------------------------------------------------------------------  Test if low_alt < p.z < high_alt
def cbt_match_alt(p)
if p.z >= $cbt_low_alt and p.z <= $cbt_high_alt
  return true
  else
  return false
end
end
#-----------------------------------------------------------------------------  Test if hit object matches layer mask
def cbt_match_layer(hit_objet)
return true if $cbt_layer_mask == ""
case hit_objet[1][0].typename
  when "Face"
    f = hit_objet[1][0]
  when "ComponentInstance"
    f = hit_objet[1][1]
  when "Group"
    f = hit_objet[1][1]
  when "Image"
    f = hit_objet[1][1]
  when "Edge"
    f = hit_objet[1][0]
end

if f.layer.name == $cbt_layer_mask
  return true
  else
  return false
end
end
#----------------------------------------------------------------------------- Returns array of comps definitions
def cbt_load_comps(comps_array)
  $cbt_comps=[]
  Sketchup.active_model.definitions.each do |d|
    if comps_array.include? d.name
      $cbt_comps.push d
    end
  end
  return $cbt_comps
end
#----------------------------------------------------------------------------- Returns number of comps to spray
def cbt_n_points(shape,center,radius,pressure,shape_points)
  press = pressure.split("%")[0].to_i 
  case shape
    when "Line"
      w = cbt_average_widths($cbt_comps)
    when "Circle"
      area = $cbt_radius * $cbt_radius * Math::PI
    when "Square"
      area = (($cbt_radius / Math.sqrt(2))*2) * (($cbt_radius / Math.sqrt(2))*2)
    when "Triangle"
      area = ($cbt_radius * $cbt_radius) * 3.0
    when "Rectangle"
      area = (shape_points[0].distance shape_points[1]) * (shape_points[1].distance shape_points[2])
    when "Selection"
      area = cbt_faces_area(cbt_sort_faces(Sketchup.active_model.selection))
  end

  case shape
    when "Line"
      n = (center.distance(radius) / w)
    else
      average_area = cbt_average_bounds($cbt_comps)
      n = area / average_area
  end

    puts shape + "   " + pressure + "   " + n.to_i.to_s + " -> " + ((n.to_i * press) / 100).floor.to_s
    cbt_warning((n.to_i * press) / 100)
if (n.to_i * press) / 100 < 1
  return 1
  else
  return (n.to_i * press) / 100
end
end
#----------------------------------------------------------------------------- Returns random number: low_range < number < high_range
def cbt_rand_range(low,high)
  v1 = rand()* (high.to_f.abs - low.to_f.abs)
  v2 = rand()* (high.to_f.abs - low.to_f.abs)
  #v3 = rand()* (high.to_f.abs - low.to_f.abs)
  if $cbt_proportion == "Yes"
    v2 = v1
  end
  if low.to_f < 0 or high.to_f < 0
    return [neg_rand(low.to_f.abs + v1),(low.to_f.abs + v2),(low.to_f.abs + v1)]
    else
    return [(low.to_f.abs + v1),(low.to_f.abs + v2),(low.to_f.abs + v1)]
  end
end
#----------------------------------------------------------------------------- Returns either x or -x, 1/2 probability
def neg_rand(x)
if rand() < 0.5
  return x
  else
  return -x
end
end
#----------------------------------------------------------------------------- Problem ?
def cbt_control_stop(trials_done, n_done, n_target)
if (n_done.to_f + 1)/trials_done.to_f < $cbt_ratio # default 0.001
  rep = UI.messagebox((trials_done-1).to_s + " trials to drop components done.\nMaybe there is no surface on which to drop,\nor there are too much constraints.\nContinue ?", MB_YESNO)
  if rep == 7
    break
    else
    $cbt_ratio = $cbt_ratio / 2.0
  end
end

if $cbt_stop_timer == true
  rep = UI.messagebox("More than " + ($cbt_delay / 60).to_s + " minutes remaining to complete.\nContinue ?", MB_YESNO)
  if rep == 7
    break
    else
    $cbt_delay = $cbt_delay * 2
    $cbt_stop_timer = false
  end
end
end
#----------------------------------------------------------------------------- Collect all components, returns array
def cbt_collect_components_old
model=Sketchup.active_model
comps = model.definitions
comps_names = []
comps.each do |c|
  if not c.group? and not c.image?
    comps_names.push(c.name)
  end
end
return comps_names.sort!
end

def cbt_collect_components
model=Sketchup.active_model
comps = model.definitions
comps_defs = []
comps.each do |c|
  if not c.group? and not c.image?
    comps_defs.push(c)
  end
end
return comps_defs
end
#----------------------------------------------------------------------------- Collect level 1 components, returns array
def cbt_collect_top_components_old
Sketchup.set_status_text("Collecting components...")
model=Sketchup.active_model
all_comps = model.definitions
all_comps_names = cbt_collect_components
top_comps = all_comps_names
all_comps.each do |cd|
  if not cd.image? and not cd.group? and not cd.internal?
    # Search for sub-components in this comp def
    cd.entities.each do |e|
      if e.typename == "ComponentInstance" #or e.typename == "Group"
        top_comps.delete e.definition.name if e.definition.internal? 
      end
    end
  end
end
Sketchup.set_status_text("")
return top_comps.sort!
end

def cbt_collect_top_components
Sketchup.set_status_text("Collecting components...")
model=Sketchup.active_model
all_comps = model.definitions
all_comps_defs = cbt_collect_components
top_comps=[]
all_comps_defs.each do |cd|
  top_comps.push(cd.name) if not cd.hidden?
end
Sketchup.set_status_text("")
return top_comps.sort!
end

#----------------------------------------------------------------------------- Collect layers names, returns array
def cbt_collect_layers
Sketchup.set_status_text("Collecting layers...")
model=Sketchup.active_model
layers_names = []
model.layers.each { |l| layers_names.push(l.name) }
Sketchup.set_status_text("")
return layers_names.sort!
end

def cbt_sort_faces(s)
Sketchup.set_status_text("Sorting faces...")
f=[]
s.each { |obj| f.push(obj) if obj.typename == "Face" }
return f
end
#----------------------------------------------------------------------------- Eradicate non-faces objects through selection
def clean_sel
Sketchup.set_status_text("Sorting selection...")
s = Sketchup.active_model.selection
s.each { |e| s.remove(e) if not e.kind_of?(Sketchup::Face) }
end
#----------------------------------------------------------------------------- Cast ray through point p and return [intersection,face] with faces set, if any
def cbt_raycast_faces(p,faces)
hit = false
hit_face = nil
v = Geom::Vector3d.new(0, 0, -100000)
faces.each do |face|
  raycast = Geom.intersect_line_plane([p,p.offset(v)], face.plane)
  if raycast and Geom.point_in_polygon_2D(p, face.vertices, true)
    hit = true
    hit_face = face
    return [Geom.intersect_line_plane([p,p.offset(v)], hit_face.plane), hit_face]
  end
end
if hit == true
  return [Geom.intersect_line_plane([p,p.offset(v)], hit_face.plane), hit_face]
else
  return false
end
end
#----------------------------------------------------------------------------- Modify comp transform, depending on options
def cbt_rotate_scale(h,c)
# Rotation
if $cbt_rotate == "Yes"
  o=c.transformation.origin
  if $cbt_perpendicular == "No"
    if h[1][0].class.to_s == "Sketchup::Face"
      r = Geom::Transformation.new([0,0,0], h[1][0].normal)
      c.transformation = c.transformation * r
    else
      r = Geom::Transformation.new([0,0,0],h[1][1].normal)
      c.transformation = c.transformation * r
    end
  else
    r=Geom::Transformation.rotation(o,[0,0,1],rand(Math::PI*2.0).to_f)
    c.transform! r
  end
end
if $cbt_low_scale != 1.0 or $cbt_high_scale != 1.0
  scales = cbt_rand_range($cbt_low_scale,$cbt_high_scale)
  c.transformation = c.transformation * Geom::Transformation.scaling(scales[0],scales[1], scales[0].abs)
end
end
#----------------------------------------------------------------------------- Returns an array of n random points based on shape
def cbt_random_points(n, shape, center, corner)
  model = Sketchup.active_model
  sel = Sketchup.active_model.selection
  v = Geom::Vector3d.new(0,0,-10000)
  rand_pts=[]

  # Set Z of random points to highest Z altitude of model + 1000 units (allow stacking higher than model top)
  z = cbt_model_top + 100000
  center = [center.x,center.y, z]
  corner = [corner.x,corner.y, z] if shape != "Rectangle" and shape != "Selection"
  p1_line = center if shape == "Line"
  p2_line = corner if shape == "Line"
  @pb = ProgressBar.new(n) 
  count = 0
  trials = 0
  
  # Layer management
  if $cbt_layer != ""
    old_layer = model.active_layer
    model.active_layer = $cbt_layer
  end
  
  case shape
    when "Selection"
      pts = map_z(corner, z)
      center = [center.x,center.y, z]
      faces = cbt_sort_faces(sel)
      while count < n
        p = polar(center, Geom::Point3d.new($cbt_radius,center.y,center.z), rand($cbt_radius), rand*(Math::PI*2))
        hit =  model.raytest [p,v]
        trials = trials + 1
        if hit != nil and Geom.point_in_polygon_2D(hit[0], pts, true) and cbt_raycast_faces(p,faces) and hit[1][0].class.to_s != $cbt_prevent and cbt_match_slope(hit) and cbt_match_alt(hit[0]) and cbt_match_layer(hit)
          comp = Sketchup.active_model.entities.add_instance $cbt_comps[rand($cbt_comps.length)], hit[0]
          # Rotate and scale
          cbt_rotate_scale(hit, comp)
           
          count = count + 1
        end
        @pb.update(count)
        cbt_control_stop(trials, count, n)
      end
    when "Line"
      vline = p1_line.vector_to(p2_line)
      while count < n
        vline.length = rand(p1_line.distance(p2_line)).to_f
        p = p1_line.offset(vline)
        hit =  model.raytest [p,v]
        trials = trials + 1
        if hit != nil and hit[1][0].class.to_s != $cbt_prevent and cbt_match_slope(hit) and cbt_match_alt(hit[0]) and cbt_match_layer(hit)
          comp = Sketchup.active_model.entities.add_instance $cbt_comps[rand($cbt_comps.length)], hit[0]
          # Rotate and scale
          cbt_rotate_scale(hit, comp)

          count = count + 1
        end
        @pb.update(count)
        cbt_control_stop(trials, count, n)
      end
    when "Rectangle"
      pts = map_z(corner, z)
      center = [center.x,center.y, z]
      while count < n
        p = polar(center, Geom::Point3d.new($cbt_radius,center.y,center.z), rand($cbt_radius), rand*(Math::PI*2))
        hit =  model.raytest [p,v]
        trials = trials + 1
        if hit != nil and Geom.point_in_polygon_2D(hit[0], pts, true) and hit[1][0].class.to_s != $cbt_prevent and cbt_match_slope(hit) and cbt_match_alt(hit[0]) and cbt_match_layer(hit)
          comp = Sketchup.active_model.entities.add_instance $cbt_comps[rand($cbt_comps.length)], hit[0]
          # Rotate and scale
          cbt_rotate_scale(hit, comp)

          count = count + 1
          @pb.update(count)
        end
        cbt_control_stop(trials, count, n)
      end
    when "Circle"
      ang = (Math::PI * 2) / n
      while count < n
        p = polar(center, Geom::Point3d.new($cbt_radius,center.y,center.z), rand($cbt_radius), rand*(Math::PI*2))
        hit =  model.raytest [p,v]
        trials = trials + 1
        if hit != nil and hit[1][0].class.to_s != $cbt_prevent and cbt_match_slope(hit) and cbt_match_alt(hit[0]) and cbt_match_layer(hit)
          comp = Sketchup.active_model.entities.add_instance $cbt_comps[rand($cbt_comps.length)], hit[0]
          # Rotate and scale
          cbt_rotate_scale(hit, comp)
          
          count = count + 1
          @pb.update(count)
        end
        cbt_control_stop(trials, count, n)
      end
    when "Square"
      # Calculate virtual square
      ang = Math::PI*2/4.0
      pts=[corner] 
      1.upto(3) do |i|
        p = corner.transform(Geom::Transformation.rotation(center,[0,0,1],i*(Math::PI/2.0)))
        pts.push p
      end
      while count < n
        # Create random point
        p = polar(center, Geom::Point3d.new($cbt_radius,center.y,center.z), rand($cbt_radius), rand*(Math::PI*2))
        # Add it if it is in virtual polygon
        if Geom.point_in_polygon_2D p, pts, true
          hit =  model.raytest [p,v]
          trials = trials + 1
          if hit != nil and hit[1][0].class.to_s != $cbt_prevent and cbt_match_slope(hit) and cbt_match_alt(hit[0]) and cbt_match_layer(hit)
            comp = Sketchup.active_model.entities.add_instance $cbt_comps[rand($cbt_comps.length)], hit[0]
            # Rotate and scale
            cbt_rotate_scale(hit, comp)
            
            count = count + 1
          end
          @pb.update(count)
        end
        cbt_control_stop(trials, count, n)
      end
    when "Triangle"
      # Calculate virtual triangle
      ang = Math::PI*2/3.0
      pts=[corner] 
      1.upto(2) do |i|
        p = corner.transform(Geom::Transformation.rotation(center,[0,0,1],i*(Math::PI*2.0/3.0)))
        pts.push p
      end
      count = 0
      while count < n
        # Create random point
        p = polar(center, Geom::Point3d.new($cbt_radius,center.y,center.z), rand($cbt_radius), rand*(Math::PI*2))
        # Add it if it is in virtual polygon
        if Geom.point_in_polygon_2D p, pts, true
          hit =  model.raytest [p,v]
          trials = trials + 1
          if hit != nil and hit[1][0].class.to_s != $cbt_prevent and cbt_match_slope(hit) and cbt_match_alt(hit[0]) and cbt_match_layer(hit)
            comp = Sketchup.active_model.entities.add_instance $cbt_comps[rand($cbt_comps.length)], hit[0]
            # Rotate and scale
            cbt_rotate_scale(hit, comp)

          count = count + 1
          @pb.update(count)
        end # if hit OK
        end
        
      end
      cbt_control_stop(trials, count, n)
  end
  
  # Layer management
  if $cbt_layer != ""
    model.active_layer = old_layer
  end
  
  return rand_pts
end
#----------------------------------------------------------------------------- Return point @ dist,angle
def polar(p1, p2, d, angle)
  v = p2 - p1
  v.length = d
  translate = Geom::Transformation.translation(v)
  p_temp = p1.transform(translate)
  rotate = Geom::Transformation.rotation( p1, Geom::Vector3d.new(0, 0, 1), angle)
  #result = p_temp.transform(rotate)
  #return result
  return p_temp.transform(rotate)
end
#----------------------------------------------------------------------------- Main dialog: prompt for brush style
def cbt_dialog_main
  model = Sketchup.active_model
  sel = model.selection
  comps = cbt_collect_top_components
  if comps.length==0
    UI.messagebox("No components loaded in this model.")
    return nil
  end
  comps.unshift("")
  l1_comps = comps.join("|")
  l2_comps = comps.join("|")
  layers = cbt_collect_layers
  Sketchup.set_status_text("Dialog box...")
  
  shapes = ["Point", "Line", "Vertices of faces", "Triangle", "Square", "Rectangle", "Circle", "Spray"]
  # Case selection
  if not sel.empty? and cbt_sort_faces(sel) != []
    shapes.push "Selection"
    $cbt_shape = "Selection"
  end
  if $cbt_shape == "Selection" and sel.empty?
    $cbt_shape = "Circle"
  end
  # Trials ratio and stop_timer
  $cbt_ratio = 0.001
  $cbt_stop_timer = false
  $cbt_delay = 180
  
  $cbt_use_settings = "Dialog" if not $cbt_use_settings
  pressures = ["5%","10%","15%","20%","25%","30%","35%","40%","45%","50%","55%","60%","65%","70%","75%","80%","85%","90%","95%","100%"]
  $cbt_name_1 = comps[0] if not $cbt_name_1 and comps.length > 0
  $cbt_name_2 = "" if not $cbt_name_2
  $cbt_name_3 = "" if not $cbt_name_3
  $cbt_name_4 = "" if not $cbt_name_4
  $cbt_name_5 = "" if not $cbt_name_5
  $cbt_name_6 = "" if not $cbt_name_6
  $cbt_shape = "Circle" if not $cbt_shape
  $cbt_radius = 50.m if not $cbt_radius
  $cbt_spray_radius = 10.m if not $cbt_spray_radius
  $cbt_pressure = "50%" if not $cbt_pressure
  $cbt_rotate = "Yes" if not $cbt_rotate
  $cbt_stacking = "Yes" if not $cbt_stacking
  $cbt_perpendicular = "Yes" if not $cbt_perpendicular
  $cbt_layer = model.active_layer.name if not $cbt_layer
  $cbt_layer_mask = "" if not $cbt_layer_mask
  $cbt_low_scale = 1.0 if not $cbt_low_scale
  $cbt_high_scale = 1.0 if not $cbt_high_scale
  $cbt_proportion = "Yes" if not $cbt_proportion
  $cbt_low_steep = 0 if not $cbt_low_steep
  $cbt_high_steep = 90 if not $cbt_high_steep
  $cbt_low_alt = cbt_model_bottom
  $cbt_high_alt = cbt_model_top
  
  prompts = ["Use settings: ", "Component 1: ", "Component 2: ", "Component 3: ", "Component 4: ", "Component 5: ", "Component 6: ", "Spray shape: ", "Spray pressure: ", "Random rotate: ", "Prevent stacking: ", "Keep vertical: ", "Destination layer: ", "Layer: mask: ", "Maintain proportions: ", "Low scale value: ", "High scale value: ", "Low steepness: ", "High steepness: " , "Low altitude: ", "High altitude: "]
  enums = [["Basic","Defaults","Dialog"].join("|"),l1_comps,l2_comps,l2_comps,l2_comps,l2_comps,l2_comps,shapes.join("|"), pressures.join("|"), ["Yes", "No"].join("|"), ["Yes", "No"].join("|"), ["Yes", "No"].join("|"), layers.unshift("").join("|"), layers.unshift("").join("|"), ["Yes", "No"].join("|")]
  values = [$cbt_use_settings, $cbt_name_1, $cbt_name_2, $cbt_name_3, $cbt_name_4, $cbt_name_5, $cbt_name_6, $cbt_shape, $cbt_pressure, $cbt_rotate, $cbt_stacking, $cbt_perpendicular, $cbt_layer, $cbt_layer_mask, $cbt_proportion, $cbt_low_scale, $cbt_high_scale, $cbt_low_steep, $cbt_high_steep, $cbt_low_alt, $cbt_high_alt]
  results = inputbox prompts, values, enums, (comps.length-1).to_s + " components available:"
  return if not results
  $cbt_use_settings, $cbt_name_1, $cbt_name_2, $cbt_name_3, $cbt_name_4, $cbt_name_5, $cbt_name_6, $cbt_shape, $cbt_pressure, $cbt_rotate, $cbt_stacking, $cbt_perpendicular, $cbt_layer, $cbt_layer_mask, $cbt_proportion, $cbt_low_scale, $cbt_high_scale, $cbt_low_steep, $cbt_high_steep, $cbt_low_alt, $cbt_high_alt = results
  
  # Load components definitions
  $cbt_comps = cbt_load_comps([$cbt_name_1, $cbt_name_2, $cbt_name_3, $cbt_name_4, $cbt_name_5, $cbt_name_6])
  
  # Layer
  if $cbt_layer == nil or $cbt_layer == ""
    $cbt_layer=model.active_layer.name
  end
  
  # Stacking
  if $cbt_stacking == "Yes"
    $cbt_prevent = "Sketchup::ComponentInstance"
    else
    $cbt_prevent = ""
  end
  
  # Altitude
  if $cbt_low_alt > $cbt_high_alt
    $cbt_high_alt = $cbt_low_alt
    $cbt_low_alt = 0
  end
  
  # Steepness
  if $cbt_low_steep > $cbt_high_steep
    $cbt_high_steep = $cbt_low_steep
    $cbt_low_steep = 0
  end
  
  # Store current dialog settings
  cbt_store_dialog_settings
  
  # Use correct settings
  case $cbt_use_settings
    when "Basic"
      cbt_use_basic
    when "Defaults"
      cbt_use_default
  end

end

#----------------------------------------------------------------------------- Set default settings dialog
def cbt_dialog_set_defaults
  model = Sketchup.active_model
  if Sketchup.active_model.path.length==0
    UI.messagebox("Please save your model first.")
    return nil
  end
  ud_shapes = ["Point", "Line", "Vertices of faces", "Triangle", "Square", "Rectangle", "Circle", "Selection"]
  ud_pressures = ["10%","20%","30%","40%","50%","60%","70%","80%","90%","100%"]
  # Trials ratio
  $cbt_ratio = 0.001
  
  $cbt_de_shape = "Circle" if not $cbt_de_shape
  $cbt_de_radius = 50.m if not $cbt_de_radius
  $cbt_de_spray_radius = 10.m if not $cbt_de_spray_radius
  $cbt_de_pressure = "50%" if not $cbt_de_pressure
  $cbt_de_rotate = "Yes" if not $cbt_de_rotate
  $cbt_de_stacking = "Yes" if not $cbt_de_stacking
  $cbt_de_perpendicular = "Yes" if not $cbt_de_perpendicular
  $cbt_de_proportion = "Yes" if not $cbt_de_proportion
  $cbt_de_low_scale = 1.0 if not $cbt_de_low_scale
  $cbt_de_high_scale = 1.0 if not $cbt_de_high_scale
  $cbt_de_low_steep = 0 if not $cbt_de_low_steep
  $cbt_de_high_steep = 90 if not $cbt_de_high_steep
  $cbt_de_low_alt = cbt_model_bottom
  $cbt_de_high_alt = cbt_model_top
  
  prompts = ["Spray shape: ", "Spray pressure: ", "Random rotate: ", "Prevent stacking: ", "Keep vertical: ", "Maintain proportions: ", "Low scale value: ", "High scale value: ", "Low steepness: ", "High steepness:", "Low altitude: ", "High altitude: "]
  enums = [ud_shapes.join("|"), ud_pressures.join("|"), ["Yes", "No"].join("|"), ["Yes", "No"].join("|"), ["Yes", "No"].join("|"), ["Yes", "No"].join("|")]
  values = [$cbt_de_shape, $cbt_de_pressure, $cbt_de_rotate, $cbt_de_stacking, $cbt_de_perpendicular, $cbt_de_proportion, $cbt_de_low_scale, $cbt_de_high_scale, $cbt_de_low_steep, $cbt_de_high_steep, $cbt_de_low_alt, $cbt_de_high_alt]
  results = inputbox prompts, values, enums, "Defaults settings"
  return if not results
  $cbt_de_shape, $cbt_de_pressure, $cbt_de_rotate, $cbt_de_stacking, $cbt_de_perpendicular, $cbt_de_proportion, $cbt_de_low_scale, $cbt_de_high_scale, $cbt_de_low_steep, $cbt_de_high_steep, $cbt_de_low_alt, $cbt_de_high_alt = results
  # Stacking
  if $cbt_de_stacking == "Yes"
    $cbt_de_prevent = "Sketchup::ComponentInstance"
    else
    $cbt_de_prevent = ""
  end
end
#----------------------------------------------------------------------------- Set basic and default settings
def cbt_set_basic
  model = Sketchup.active_model
  if Sketchup.active_model.path.length==0
    UI.messagebox("Please save your model first.")
    return nil
  end
  # Trials ratio
  $cbt_ratio = 0.001
  # Spray radius default
  $cbt_spray_radius = 10.m
  # Basic
  $cbt_ba_shape = "Circle"
  $cbt_ba_radius = 50.m
  $cbt_ba_spray_radius = 10.m
  $cbt_ba_pressure = "50%"
  $cbt_ba_rotate = "Yes"
  $cbt_ba_stacking = "Yes"
  $cbt_ba_perpendicular = "Yes"
  $cbt_ba_proportion = "Yes"
  $cbt_ba_low_scale = 1.0
  $cbt_ba_high_scale = 1.0
  $cbt_ba_low_steep = 0
  $cbt_ba_high_steep = 90
  $cbt_ba_low_alt = cbt_model_bottom
  $cbt_ba_high_alt = cbt_model_top
  $cbt_ba_prevent = "Sketchup::ComponentInstance"
  # Default
  $cbt_de_shape = "Circle" if not $cbt_de_shape
  $cbt_de_radius = 50.m if not $cbt_de_radius
  $cbt_de_spray_radius = 10.m if not $cbt_de_spray_radius
  $cbt_de_pressure = "50%" if not $cbt_de_pressure
  $cbt_de_rotate = "Yes" if not $cbt_de_rotate
  $cbt_de_stacking = "Yes" if not $cbt_de_stacking
  $cbt_de_perpendicular = "Yes" if not $cbt_de_perpendicular
  $cbt_de_proportion = "Yes" if not $cbt_de_proportion
  $cbt_de_low_scale = 1.0 if not $cbt_de_low_scale
  $cbt_de_high_scale = 1.0 if not $cbt_de_high_scale
  $cbt_de_low_steep = 0 if not $cbt_de_low_steep
  $cbt_de_high_steep = 90 if not $cbt_de_high_steep
  $cbt_de_low_alt = cbt_model_bottom
  $cbt_de_high_alt = cbt_model_top
  $cbt_de_prevent = "Sketchup::ComponentInstance"
end
#----------------------------------------------------------------------------- Overwrite current settings with basic settings
def cbt_use_basic
  #puts "settings to basic"
  $cbt_shape = $cbt_ba_shape
  $cbt_radius = $cbt_ba_radius
  $cbt_spray_radius = $cbt_ba_spray_radius
  $cbt_pressure = $cbt_ba_pressure
  $cbt_rotate = $cbt_ba_rotate
  $cbt_stacking = $cbt_ba_stacking
  $cbt_perpendicular = $cbt_ba_perpendicular
  $cbt_proportion = $cbt_ba_proportion
  $cbt_low_scale = $cbt_ba_low_scale
  $cbt_high_scale = $cbt_ba_high_scale
  $cbt_low_steep = $cbt_ba_low_steep
  $cbt_high_steep = $cbt_ba_high_steep
  $cbt_low_alt = cbt_model_bottom
  $cbt_high_alt = cbt_model_top
  $cbt_prevent = "Sketchup::ComponentInstance"
end
#----------------------------------------------------------------------------- Overwrite current settings with basic settings
def cbt_use_default
  #puts "settings to default"
  $cbt_shape = $cbt_de_shape
  $cbt_radius = $cbt_de_radius
  $cbt_spray_radius = $cbt_de_spray_radius
  $cbt_pressure = $cbt_de_pressure
  $cbt_rotate = $cbt_de_rotate
  $cbt_stacking = $cbt_de_stacking
  $cbt_perpendicular = $cbt_de_perpendicular
  $cbt_proportion = $cbt_de_proportion
  $cbt_low_scale = $cbt_de_low_scale
  $cbt_high_scale = $cbt_de_high_scale
  $cbt_low_steep = $cbt_de_low_steep
  $cbt_high_steep = $cbt_de_high_steep
  $cbt_low_alt = cbt_model_bottom
  $cbt_high_alt = cbt_model_top
  $cbt_prevent = "Sketchup::ComponentInstance"
end
#----------------------------------------------------------------------------- Temp save of dialog settings, to be restored when command ends
def cbt_store_dialog_settings
  $cbt_t_shape = $cbt_shape
  $cbt_t_radius = $cbt_radius
  $cbt_t_pressure = $cbt_pressure
  $cbt_t_rotate = $cbt_rotate
  $cbt_t_stacking = $cbt_stacking
  $cbt_t_perpendicular = $cbt_perpendicular
  $cbt_t_proportion = $cbt_proportion
  $cbt_t_low_scale = $cbt_low_scale
  $cbt_t_high_scale = $cbt_high_scale
  $cbt_t_low_steep = $cbt_low_steep
  $cbt_t_high_steep = $cbt_high_steep
  $cbt_t_low_alt = $cbt_low_alt
  $cbt_t_high_alt = $cbt_high_alt
  $cbt_t_prevent = $cbt_prevent
end

#----------------------------------------------------------------------------- Restore dialog settings
def cbt_restore_dialog_settings
  $cbt_shape = $cbt_t_shape
  $cbt_radius = $cbt_t_radius
  $cbt_pressure = $cbt_t_pressure
  $cbt_rotate = $cbt_t_rotate
  $cbt_stacking = $cbt_t_stacking
  $cbt_perpendicular = $cbt_t_perpendicular
  $cbt_proportion = $cbt_t_proportion
  $cbt_low_scale = $cbt_t_low_scale
  $cbt_high_scale = $cbt_t_high_scale
  $cbt_low_steep = $cbt_t_low_steep
  $cbt_high_steep = $cbt_t_high_steep
  $cbt_low_alt = $cbt_t_low_alt
  $cbt_high_alt = $cbt_t_high_alt
  $cbt_prevent = $cbt_t_prevent
end
#----------------------------------------------------------------------------- Warning: lot of components !
def cbt_warning(n)
return if n < 500
n2 = n/100 * 100
rep = UI.messagebox("About to create more than " + n2.to_s + " components.\n Proceed ?",MB_YESNO)
break if rep == 7
end
#----------------------------------------------------------------------------- Tool class for Ngons
class ComponentNgonBrushTool

def initialize
  model = Sketchup.active_model
  entities = model.entities

  @ip1 = nil
  @ip2 = nil
  @ip3 = nil
  @xdown = 0
  @ydown = 0
  @state = 0
  @shape = $cbt_shape
  #puts "Init NgonTool with: " + $cbt_shape
end

def activate
    @ip1 = Sketchup::InputPoint.new
    @ip2 = Sketchup::InputPoint.new
    @ip3 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false
    # This sets the label for the VCB
    Sketchup::set_status_text "Radius or Distance", SB_VCB_LABEL
    self.reset(nil)
    #puts "Activate NgonTool with: " + $cbt_shape
end

def onMouseLeave(view)
  Sketchup.send_action "selectSelectionTool:"
end

def deactivate(view)
    view.invalidate if @drawn
end

def onMouseMove(flags, x, y, view)
    if( @state == 0 )
      #puts"state 0: " + $cbt_shape
        @ip.pick view, x, y
        if( @ip != @ip1 )
            view.invalidate if( @ip.display? or @ip1.display? )
            @ip1.copy! @ip
            # set the tooltip that should be displayed to this point
            view.tooltip = @ip1.tooltip
        end
    else
        # Getting the radius point or the second point for rectangle
        @ip2.pick view, x, y, @ip1
        view.tooltip = @ip2.tooltip if( @ip2.valid? )
        view.invalidate
        # Update the length displayed in the VCB
        if( @ip2.valid? )
            @length = @ip1.position.distance(@ip2.position)
            Sketchup::set_status_text @length.to_s, SB_VCB_VALUE
        end
        # Check to see if the mouse was moved far enough to create a circle.
        # This is used so that you can create a circle by either draggin or doing click-move-click
        if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
            @dragging = true
            self.draw_geometry(@ip1.position, @ip2.position, @length, view)
        end
    end
end

def onLButtonDown(flags, x, y, view)
    # When the user clicks the first time, we switch to getting the
    # second point.  When he clicks a second time we create the circle
    if( @state == 0 )
        @ip1.pick view, x, y
        if( @ip1.valid? )
            @state = 1
            Sketchup::set_status_text "Radius", SB_PROMPT
            @xdown = x
            @ydown = y
        end
    end
    if( @state == 1 )
        # create the ngon on the second click
        #if( @ip2.valid? and ($cbt_shape == "Circle" or $cbt_shape == "Square" or $cbt_shape == "Triangle"))
        if( @ip2.valid? and (@shape == "Circle" or @shape == "Square" or @shape == "Triangle"))
          @state = 2
          # spray components
          self.create_geometry(@ip1.position, @ip2.position,view)
          self.reset(view)
        end
    end
    # Clear any inference lock
    view.lock_inference
end

def onLButtonUp(flags, x, y, view)
    # If we are doing a drag, then create the circle on the mouse up event
    if( @dragging && @ip2.valid? )
        # spray components)
        self.create_geometry(@ip1.position, @ip2.position,view)
        self.reset(view)
    end
end

def onKeyDown(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
            view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
            view.lock_inference @ip2, @ip1
        end
    end
end

def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end

def onUserText(text, view)
    return if not @state == 1
    return if not @ip2.valid?
    
    begin
        value = text.to_l
    rescue
        # Error parsing the text
        UI.beep
        puts "Cannot convert #{text} to a Length"
        value = nil
        Sketchup::set_status_text "", SB_VCB_VALUE
    end
    return if !value

    # Compute the direction and the second point
    pt1 = @ip1.position
    vec = @ip2.position - pt1
    if( vec.length == 0.0 )
        return
    end
    vec.length = value
    pt2 = pt1 + vec

    # spray components
    self.create_geometry(pt1, pt2, view)
    self.reset(view)
end

# The draw method is called whenever the view is refreshed.
def draw(view)
    if( @ip1.valid? )
        if( @ip1.display? )
            @ip1.draw(view)
            @drawn = true
        end
        if( @ip2.valid? )
            @ip2.draw(view) if( @ip2.display? )
            view.set_color_from_line(@ip1, @ip2)
            self.draw_geometry(@ip1.position, @ip2.position, @length, view)
            @drawn = true
        end
    end
end

# onCancel is called when the user hits the escape key
def onCancel(flag, view)
    self.reset(view)
    Sketchup.send_action "selectSelectionTool:"
end

# Reset the tool back to its initial state
def reset(view)
    @state = 0
    # Display a prompt on the status bar
    Sketchup::set_status_text("Center of circle", SB_PROMPT)
    # clear the InputPoints
    @ip1.clear
    @ip2.clear
    if( view )
        view.tooltip = nil
        view.invalidate if @drawn
    end
    @drawn = false
    @dragging = false
end

# Create new geometry when the user has selected two points.
def create_geometry(p1, p2, view)
  model = Sketchup.active_model
  model.start_operation "Component Spray (" + @shape + ")"
  $cbt_radius = p1.distance p2
  #n_comps = cbt_n_points($cbt_shape, @ip1.position, $cbt_radius, $cbt_pressure,nil)
  #cbt_random_points(n_comps, $cbt_shape, @ip1.position, @ip2.position)
  n_comps = cbt_n_points(@shape, @ip1.position, $cbt_radius, $cbt_pressure,nil)
  cbt_random_points(n_comps, @shape, @ip1.position, @ip2.position)
  model.commit_operation
    
end #of def create_geometry


# Draw the geometry
def draw_geometry(pt1, pt2, radius, view)
  case @shape
    when "Circle"
      sides = 30
    when "Square"
      sides = 4
    when "Triangle"
      sides = 3
  end
  ang = Math::PI*2/sides
  pts=[pt2]
  0.upto(sides-1) do |i|
    trans = Geom::Transformation.rotation(pt1,[0,0,1],i*ang)
    pts.push pt2.transform(trans)
  end
  # Ngon temp draw
  view.draw_points(pt1,8,5,"red")
  pts.push pts.first
  view.draw_polyline pts
end

end # class ComponentNgonBrushTool

#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------

class ComponentPointBrushTool

def initialize
    @ip1 = nil
    @xdown = 0
    @ydown = 0
    @shape = $cbt_shape
    # Iterate through comp list
    @indice = 0
end

def activate
    @ip1 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false
    Sketchup::set_status_text("Point", SB_VCB_LABEL)
    self.reset(nil)
end

def onMouseLeave(view)
  Sketchup.send_action "selectSelectionTool:"
end

def deactivate(view)
    view.invalidate if @drawn
end

def onMouseMove(flags, x, y, view)
  @ip.pick view, x, y
  if( @ip != @ip1 )
    view.invalidate if( @ip.display? or @ip1.display? )
    @ip1.copy! @ip
    view.tooltip = @ip1.tooltip
  end
end

def onLButtonDown(flags, x, y, view)
        @ip1.pick view, x, y
        if( @ip1.valid? )
            Sketchup::set_status_text("Point", SB_PROMPT)
            @xdown = x
            @ydown = y
        end
    # Clear any inference lock
    view.lock_inference
    self.create_geometry(@ip1.position,view)
end

def onLButtonUp(flags, x, y, view)
  if( @dragging && @ip2.valid? )
    self.reset(view)
  end
end

def onKeyDown(key, repeat, flags, view)
  if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
    @shift_down_time = Time.now
    # if we already have an inference lock, then unlock it
    if( view.inference_locked? )
      # calling lock_inference with no arguments actually unlocks
      view.lock_inference
    end
  end
end

def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end

def onUserText(text, view)
    self.reset(view)
end

def draw(view)
  if( @ip1.valid? )
    if( @ip1.display? )
      @ip1.draw(view)
      @drawn = true
    end
  end
end

def onCancel(flag, view)
    self.reset(view)
    Sketchup.send_action "selectSelectionTool:"
end

def reset(view)
  @state = 0
  Sketchup::set_status_text($exStrings.GetString("Point"), SB_PROMPT)
  @ip1.clear
  if( view )
    view.tooltip = nil
    view.invalidate if @drawn
  end
  @drawn = false
  @dragging = false
end

def create_geometry(p1, view)
  model = Sketchup.active_model
  model.start_operation "Component Spray (Point)"
  z = cbt_model_top + 100000
  p = [@ip1.position.x, @ip1.position.y, z]
  v = Geom::Vector3d.new(0,0,-10000)
  hit =  model.raytest [p,v]
  
  if hit != nil and hit[1][0].class.to_s != $cbt_prevent and cbt_match_slope(hit) and cbt_match_alt(hit[0])
    
    # Random type
    #comp = Sketchup.active_model.entities.add_instance $cbt_comps[rand($cbt_comps.length)], hit[0]
    
    # Iterate through selected components
    if @indice == $cbt_comps.length
      @indice = 0
    end
    comp = Sketchup.active_model.entities.add_instance $cbt_comps[@indice], hit[0]
    @indice = @indice + 1
    
    # Rotation and scale
    cbt_rotate_scale(hit,comp)
    else
    Sketchup.set_status_text("Cannot place a component here !")
  end
  if not cbt_match_slope(hit)
      Sketchup.set_status_text("Slope out of range. Cannot place a component here !")
  end
  if not cbt_match_alt(hit[0])
      Sketchup.set_status_text("Altitude out of range. Cannot place a component here !")
  end
  
  model.commit_operation
end

def draw_geometry(pt1, view)
    view.draw_point(pt1)
end

end # class ComponentPointBrushTool

#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------

class ComponentLineBrushTool

def initialize
    @ip1 = nil
    @ip2 = nil
    @xdown = 0
    @ydown = 0
    @shape = $cbt_shape
end

def activate
    @ip1 = Sketchup::InputPoint.new
    @ip2 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false
    Sketchup::set_status_text $exStrings.GetString("Length"), SB_VCB_LABEL
    self.reset(nil)
end

def onMouseLeave(view)
  Sketchup.send_action "selectSelectionTool:"
end

def deactivate(view)
    view.invalidate if @drawn
end

def onMouseMove(flags, x, y, view)
    if( @state == 0 )
        @ip.pick view, x, y
        if( @ip != @ip1 )
            view.invalidate if( @ip.display? or @ip1.display? )
            @ip1.copy! @ip
            view.tooltip = @ip1.tooltip
        end
    else
        @ip2.pick view, x, y, @ip1
        view.tooltip = @ip2.tooltip if( @ip2.valid? )
        view.invalidate

        if( @ip2.valid? )
            length = @ip1.position.distance(@ip2.position)
            Sketchup::set_status_text length.to_s, SB_VCB_VALUE
        end
        
        if( (x-@xdown).abs > 10 || (y-@ydown).abs > 10 )
            @dragging = true
        end
    end
end

def onLButtonDown(flags, x, y, view)
    if( @state == 0 )
        @ip1.pick view, x, y
        if( @ip1.valid? )
            @state = 1
            Sketchup::set_status_text $exStrings.GetString("Select second end"), SB_PROMPT
            @xdown = x
            @ydown = y
        end
    else
        # create the line on the second click
        if( @ip2.valid? )
            self.create_geometry(@ip1.position, @ip2.position,view)
            self.reset(view)
        end
    end
    # Clear any inference lock
    view.lock_inference
end

def onLButtonUp(flags, x, y, view)
    if( @dragging && @ip2.valid? )
        self.create_geometry(@ip1.position, @ip2.position,view)
        self.reset(view)
    end
end

def onKeyDown(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
        @shift_down_time = Time.now
        
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
            # calling lock_inference with no arguments actually unlocks
            view.lock_inference
        elsif( @state == 0 && @ip1.valid? )
            view.lock_inference @ip1
        elsif( @state == 1 && @ip2.valid? )
            view.lock_inference @ip2, @ip1
        end
    end
end

def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end

def onUserText(text, view)
    return if not @state == 1
    return if not @ip2.valid?

    begin
        value = text.to_l
    rescue
        # Error parsing the text
        UI.beep
        puts "Cannot convert #{text} to a Length"
        value = nil
        Sketchup::set_status_text "", SB_VCB_VALUE
    end
    return if !value

    # Compute the direction and the second point
    pt1 = @ip1.position
    vec = @ip2.position - pt1
    if( vec.length == 0.0 )
        UI.beep
        return
    end
    vec.length = value
    pt2 = pt1 + vec

    # Create a line
    self.create_geometry(pt1, pt2, view)
    self.reset(view)
end

def draw(view)
    if( @ip1.valid? )
        if( @ip1.display? )
            @ip1.draw(view)
            @drawn = true
        end
        
        if( @ip2.valid? )
            @ip2.draw(view) if( @ip2.display? )
            view.set_color_from_line(@ip1, @ip2)
            self.draw_geometry(@ip1.position, @ip2.position, view)
            @drawn = true
        end
    end
end

def onCancel(flag, view)
    self.reset(view)
    Sketchup.send_action "selectSelectionTool:"
end

def reset(view)
    @state = 0
    Sketchup::set_status_text($exStrings.GetString("Select first end"), SB_PROMPT)
    @ip1.clear
    @ip2.clear
    
    if( view )
        view.tooltip = nil
        view.invalidate if @drawn
    end
    
    @drawn = false
    @dragging = false
end

def create_geometry(p1, p2, view)
  model = Sketchup.active_model
  model.start_operation "Component Spray (Line)"
  $cbt_radius = p1.distance p2
  n_comps = cbt_n_points(@shape, @ip1.position, @ip2.position, $cbt_pressure,nil)
  cbt_random_points(n_comps, @shape, @ip1.position, @ip2.position)
  model.commit_operation
end

def draw_geometry(pt1, pt2, view)
    view.draw_line(pt1, pt2)
end

end # class ComponentLineBrushTool


#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------

class ComponentSprayBrushTool

def initialize
    @ip1 = nil
    @xdown = 0
    @ydown = 0
    @shape = "Circle"
end

def activate
    @ip1 = Sketchup::InputPoint.new
    @ip = Sketchup::InputPoint.new
    @drawn = false
    Sketchup::set_status_text("Spray radius", SB_VCB_LABEL)
    self.reset(nil)
end

def onMouseLeave(view)
  Sketchup.send_action "selectSelectionTool:"
end

def deactivate(view)
    view.invalidate if @drawn
end

def onMouseMove(flags, x, y, view)
  @ip.pick view, x, y
  if( @ip != @ip1 )
    view.invalidate if( @ip.display? or @ip1.display? )
    @ip1.copy! @ip
    view.tooltip = @ip1.tooltip
  end
  
end

def onLButtonDown(flags, x, y, view)
        @ip1.pick view, x, y
        if( @ip1.valid? )
            Sketchup::set_status_text("Point", SB_PROMPT)
            @xdown = x
            @ydown = y
        end
    # Clear any inference lock
    view.lock_inference
    self.create_geometry(@ip1.position,view)
end

def onLButtonUp(flags, x, y, view)
  if( @dragging && @ip2.valid? )
    self.reset(view)
  end
end

def onKeyDown(key, repeat, flags, view)
  if( key == CONSTRAIN_MODIFIER_KEY && repeat == 1 )
    @shift_down_time = Time.now
    # if we already have an inference lock, then unlock it
    if( view.inference_locked? )
      # calling lock_inference with no arguments actually unlocks
      view.lock_inference
    end
  end
end

def onKeyUp(key, repeat, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end

def onUserText(text, view)
    begin
        value = text.to_l
    rescue
        # Error parsing the text
        UI.beep
        puts "Cannot convert #{text} to a radius"
        value = nil
        Sketchup::set_status_text "", SB_VCB_VALUE
    end
    return if !value
    if value
      $cbt_spray_radius = value.to_l
    end    
    self.reset(view)
end

def draw(view)
  if( @ip1.valid? )
    if( @ip1.display? )
      @ip1.draw(view)
      v_rad = Geom::Vector3d.new($cbt_spray_radius,0,0)
      self.draw_geometry(@ip1.position, @ip1.position.offset(v_rad), $cbt_spray_radius, view)
      @drawn = true
    end
  end
end

def onCancel(flag, view)
    self.reset(view)
    Sketchup.send_action "selectSelectionTool:"
end

def reset(view)
  @state = 0
  Sketchup::set_status_text($exStrings.GetString("Point"), SB_PROMPT)
  @ip1.clear
  if( view )
    view.tooltip = nil
    view.invalidate if @drawn
  end
  @drawn = false
  @dragging = false
end

def create_geometry(p1, view)
  model = Sketchup.active_model
  model.start_operation "Component Spray (Spray)"
  p2= p1.offset(Geom::Vector3d.new($cbt_spray_radius,0,0))
  $cbt_radius = $cbt_spray_radius
  n_comps = cbt_n_points(@shape,p1, $cbt_radius, $cbt_pressure, nil)
  cbt_random_points(n_comps, @shape, p1, p2)
  model.commit_operation
end

# Draw the geometry
def draw_geometry(pt1, pt2, radius, view)
  sides = 12
  ang = Math::PI*2/sides
  pts=[]
  top = [pt1.x,pt1.y, pt1.z+$cbt_spray_radius]
  
  0.upto(sides-1) do |i|
    pts.push(pt2.transform(Geom::Transformation.rotation(pt1,[0,0,1],i*ang)))
  end
  
  0.upto(sides-1) do |i|
    pts.push(pt2.transform(Geom::Transformation.rotation(pt1,[0,0,1],i*ang)))
    view.draw GL_LINES, [top, pt2.transform(Geom::Transformation.rotation(pt1,[0,0,1],i*ang))]
  end
  pts.push pts.first
  view.draw_polyline pts
  
end

end # class ComponentSprayBrushTool
#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------

class ComponentRectBrushTool

def initialize
    @ip = Sketchup::InputPoint.new
    @ip1 = Sketchup::InputPoint.new
    @shape = $cbt_shape
    reset
end

def reset
    @pts = []
    @state = 0
    @ip1.clear
    @drawn = false
    Sketchup::set_status_text "", SB_VCB_LABEL
    Sketchup::set_status_text "", SB_VCB_VALUE
    Sketchup::set_status_text "Start point"
    @shift_down_time = Time.now
end

def activate
    self.reset
end

def deactivate(view)
    view.invalidate if @drawn
end

def onMouseLeave(view)
  Sketchup.send_action "selectSelectionTool:"
end

def set_current_point(x, y, view)
    if( !@ip.pick(view, x, y, @ip1) )
        return false
    end
    need_draw = true
    
    # Set the tooltip that will be displayed
    view.tooltip = @ip.tooltip
        
    # Compute points
    case @state
    when 0
        @pts[0] = @ip.position
        @pts[4] = @pts[0]
        need_draw = @ip.display? || @drawn
    when 1
        @pts[1] = @ip.position
        @width = @pts[0].distance @pts[1]
        Sketchup::set_status_text @width.to_s, SB_VCB_VALUE
    when 2
        pt1 = @ip.position
        pt2 = pt1.project_to_line @pts
        vec = pt1 - pt2
        @height = vec.length
        if( @height > 0 )
            # test for a square
            square_point = pt2.offset(vec, @width)
            if( view.pick_helper.test_point(square_point, x, y) )
                @height = @width
                @pts[2] = @pts[1].offset(vec, @height)
                @pts[3] = @pts[0].offset(vec, @height)
                view.tooltip = "Square"
            else
                @pts[2] = @pts[1].offset(vec)
                @pts[3] = @pts[0].offset(vec)
            end
        else
            @pts[2] = @pts[1]
            @pts[3] = @pts[0]
        end
        Sketchup::set_status_text @height.to_s, SB_VCB_VALUE
    end

    view.invalidate if need_draw
end

def onMouseMove(flags, x, y, view)
    self.set_current_point(x, y, view)
end

def create_geometry
  model = Sketchup.active_model
  model.start_operation "Component Spray (Rectangle)"
  
  $cbt_radius = (@pts[0].distance @pts[2]) / 2.0
  n_comps = cbt_n_points(@shape, @pts[0], @pts[2], $cbt_pressure, @pts)
  diag = @pts[0].vector_to @pts[2]
  diag.length = (@pts[0].distance @pts[2])/2.0
  rect_center = @pts[0].offset(diag)
  @pts.pop
  cbt_random_points(n_comps, @shape, rect_center, @pts)

  model.commit_operation
  self.reset
end

def increment_state
    @state += 1
    case @state
    when 1
        @ip1.copy! @ip
        Sketchup::set_status_text "Second point"
        Sketchup::set_status_text "Length", SB_VCB_LABEL
        Sketchup::set_status_text "", SB_VCB_VALUE
    when 2
        @ip1.clear
        Sketchup::set_status_text "Third point"
        Sketchup::set_status_text "Width", SB_VCB_LABEL
        Sketchup::set_status_text "", SB_VCB_VALUE
    when 3
        self.create_geometry
    end
end

def onLButtonDown(flags, x, y, view)
    self.set_current_point(x, y, view)
    self.increment_state
    view.lock_inference
end

def onCancel(flag, view)
    view.invalidate if @drawn
    self.reset
end

# This is called when the user types a value into the VCB
def onUserText(text, view)
    begin
        value = text.to_l
    rescue
        # Error parsing the text
        UI.beep
        value = nil
        Sketchup::set_status_text "", SB_VCB_VALUE
    end
    return if !value
    
    case @state
    when 1
        # update the width
        vec = @pts[1] - @pts[0]
        if( vec.length > 0.0 )
            vec.length = value
            @pts[1] = @pts[0].offset(vec)
            view.invalidate
            self.increment_state
        end
    when 2
        # update the height
        vec = @pts[3] - @pts[0]
        if( vec.length > 0.0 )
            vec.length = value
            @pts[2] = @pts[1].offset(vec)
            @pts[3] = @pts[0].offset(vec)
            self.increment_state
        end
    end
end

def getExtents
    bb = Geom::BoundingBox.new
    case @state
    when 0
        # We are getting the first point
        if( @ip.valid? && @ip.display? )
            bb.add @ip.position
        end
    when 1
        bb.add @pts[0]
        bb.add @pts[1]
    when 2
        bb.add @pts
    end
    bb
end

def draw(view)
    @drawn = false
    
    # Show the current input point
    if( @ip.valid? && @ip.display? )
        @ip.draw(view)
        @drawn = true
    end

    # show the rectangle
    if( @state == 1 )
        # just draw a line from the start to the end point
        view.set_color_from_line(@ip1, @ip)
        inference_locked = view.inference_locked?
        view.line_width = 3 if inference_locked
        view.draw(GL_LINE_STRIP, @pts[0], @pts[1])
        view.line_width = 1 if inference_locked
        @drawn = true
    elsif( @state > 1 )
        # draw the curve
        view.drawing_color = "red"
        view.draw(GL_LINE_STRIP, @pts)
        @drawn = true
    end
end

def onKeyDown(key, rpt, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY && rpt == 1 )
        @shift_down_time = Time.now
        
        # if we already have an inference lock, then unlock it
        if( view.inference_locked? )
            view.lock_inference
        elsif( @state == 0 )
            view.lock_inference @ip
        elsif( @state == 1 )
            view.lock_inference @ip, @ip1
        end
    end
end

def onKeyUp(key, rpt, flags, view)
    if( key == CONSTRAIN_MODIFIER_KEY &&
        view.inference_locked? &&
        (Time.now - @shift_down_time) > 0.5 )
        view.lock_inference
    end
end

end # of class RectangleTool



#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------

def facesbrushtool(shape)
model = Sketchup.active_model
sel = model.selection
if not sel.empty?
  clean_sel
  else
  UI.messagebox("No selection.")
  return nil
end
model.start_operation "Component Spray (Faces set)"
bb = cbt_faces_bounds(cbt_sort_faces(model.selection))
corners = [(bb.corner 0),(bb.corner 1),(bb.corner 3),(bb.corner 2)]
# Virtual polygon of BB from which to cast rays
pts = map_z(corners, cbt_model_top+100000)
$cbt_radius = (pts[0].distance pts[2]) / 2.0
n_comps = cbt_n_points(shape, pts[0], pts[2], $cbt_pressure, pts)
diag = pts[0].vector_to pts[2]
diag.length = (pts[0].distance pts[2])/2.0
center = pts[0].offset(diag)
cbt_random_points(n_comps, shape, center, pts)

model.commit_operation
end

def verticesbrushtool(shape)
model = Sketchup.active_model
sel = model.selection
if not sel.empty?
  clean_sel
  else
  UI.messagebox("No selection.")
  return nil
end
model.start_operation "Composant Spray (Faces vertices)"
# Get unique array of vertices
vertices_array=[]
sel.each do |f|
  f.loops.each do |loop|
    loop.vertices.each do |v|
      vertices_array.push(v) if not vertices_array.include?(v) 
    end
  end
end
# Randomly drop components at vertices
vertices_array.each do |p|
  comp = Sketchup.active_model.active_entities.add_instance $cbt_comps[rand($cbt_comps.length)], p.position
  cbt_rotate_scale(p,comp)
end
model.commit_operation
end

def componentbrushtool
  # Exit current command, if any
  Sketchup.send_action "selectSelectionTool:"
  # Dialog box brush style
  cbt_dialog_main
  # Switch
  case $cbt_shape
    when "Point"
      Sketchup.active_model.select_tool ComponentPointBrushTool.new
    when "Line"
      Sketchup.active_model.select_tool ComponentLineBrushTool.new
    when "Vertices of faces"
      verticesbrushtool("Selection")
    when "Rectangle"
      Sketchup.active_model.select_tool ComponentRectBrushTool.new
    when "Selection"
      facesbrushtool("Selection")
    when "Spray"
      Sketchup.active_model.select_tool ComponentSprayBrushTool.new
    else
      Sketchup.active_model.select_tool ComponentNgonBrushTool.new
    end

  # Restore dialog setings
  cbt_restore_dialog_settings  
end


#----------------------------------------------------------------------------------------------------------
#----------------------------------------------------------------------------------------------------------

# initialize basic settings
model = Sketchup.active_model
if Sketchup.active_model.path.length!=0
  cbt_set_basic
end




#***********************************************************************************************************************************************************************************
#                       BALUSTRADE COMPONENTS
#***********************************************************************************************************************************************************************************
# Name :        Balustrade
# Description : Creates balustrade with desired dimmensions
# Author :      Tomasz MAREK, modified D. Bur
# Usage :       1.Menu - Draw/Balustrade
#		2.Give dimmensions and Ok.			
# Date :        14.Jul.2oo5, December 2005
# Type :        Architecture tool
# History:        
#                
#                 1.o  (14.July.2oo5) - first version, version 1.1 & 1.2. - November 2005
def lines2balustrade
model = Sketchup.active_model
$ss = model.selection
notanedge = 0
if $ss.empty? 
  UI.messagebox("No selection. Select lines please.")
  return nil
end
$ss.each do |e|
  if( e.typename != "Edge")
    $ss.remove e
    notanedge = notanedge + 1
  end
end
if (notanedge > 0)
  UI.messagebox( notanedge.to_s + " objects are not lines and will be ignored.")
end
# default_values
$handrail_section = "Rectangular" if not $handrail_section
$post_section = "Rectangular" if not $post_section
$panel_section = "None" if not $panel_section
$height = 1000.mm if not $height
$jump = 500.mm if not $jump
$post_width = 50.mm if not $post_width
$post_depth = 50.mm if not $post_depth
$post_diam = 50.mm if not $post_diam
$handrail_height= 40.mm if not $handrail_height
$handrail_depth = 60.mm if not $handrail_depth
$handrail_diam = 60.mm if not $handrail_diam
$panel_start_height = 200.mm if not $panel_start_height 
$panel_height = 500.mm if not $panel_height
$panel_thickness = 10.mm if not $panel_thickness
# Handrail status for followme only once
$handrail_done = false
# Available handrail section components
hr_components_list = (collect_components "hr", "HR")
hr_components_list.sort!
hr_components_list.reverse!
# Available post section components
po_components_list = (collect_components "po", "PO")
po_components_list.sort!
po_components_list.reverse!
# Available panel section components
pa_components_list = (collect_components "pa", "PA")
pa_components_list.sort!
pa_components_list.reverse!
inputbox1 = hr_components_list.join("|") + "|Circular|Rectangular"
inputbox2 = po_components_list.join("|") + "|Circular|Rectangular"
inputbox3 = pa_components_list.join("|") + "|None|Rectangular"
array_of_dropdowns = [inputbox1, inputbox2,inputbox3]
  
prompts = ["Handrail Section:  ", "Posts Section:  ", "Panels:  ", "Balustrade Height:  ", "Posts spacing:  "]
values = [$handrail_section, $post_section, $panel_section, $height, $jump]
results = inputbox prompts, values, array_of_dropdowns, "Balustrade Elements"
return if not results
$handrail_section, $post_section, $panel_section, $height, $jump = results
# Construct the prompts dialog and values
prompts = []
$values = []
$variables = []
case $handrail_section
  when "Rectangular"
    prompts.push("Handrail: width ", "Handrail: height ")
    $values.push($handrail_depth, $handrail_height)
    $variables.push("$handrail_depth", "$handrail_height")
    $handrail_diam = nil
  when "Circular"
    prompts.push("Handrail: diameter ")
    $values.push($handrail_diam)
    $variables.push("$handrail_diam")
    $handrail_height = nil
    $handrail_depth = nil
  #when "Aucune"
    #$handrail_diam = nil
    #$handrail_height = nil
    #$handrail_depth = nil
  else
    $handrail_diam = nil
    $handrail_height = nil
    $handrail_depth = nil
end
case $post_section
  when "Rectangular"
    prompts.push("Post: width ", "Post: depth ")
    $values.push($post_width, $post_depth)
    $variables.push("$post_width", "$post_depth")
    $post_diam = nil
  when "Circular"
    prompts.push("Post: diameter ")
    $values.push($post_diam)
    $variables.push("$post_diam")
    $post_width = nil
    $post_depth = nil
  #when "Aucun"
    #$post_diam = nil
    #$post_width = nil
    #$post_depth = nil
  else
    $post_diam = nil
    $post_width = nil
    $post_depth = nil
end
case $panel_section
  when "Rectangular"
    prompts.push("Panel: height above ground ", "Panel: height ", "Panel: thickness ")
    $values.push($panel_start_height, $panel_height, $panel_thickness)
    $variables.push("$panel_start_height", "$panel_height", "$panel_thickness")
  when "None"
    $panel_start_height = nil
    $panel_height = nil
    $panel_thickness = nil
  else
    prompts.push("Panel: height above ground ")
    $values.push($panel_start_height)
    $variables.push("$panel_start_height")
    if ((comp_type $panel_section) == "XY")
      prompts.push("Panel: extrusion height ")
      $values.push($panel_height)
      $variables.push("$panel_height")
      else
      $panel_height = nil
    end
    $panel_thickness = nil
end
# User inputs
if prompts!= []
  results = inputbox prompts, $values, "Balustrade dimensions"
  # Eval variables
  i = 0
  $variables.each do |v|
    case v
      when "$post_width"
        $post_width = results[i]
      when "$post_depth"
        $post_depth = results[i]
      when "$post_diam"
        $post_diam = results[i]
      when "$handrail_height"
        $handrail_height = results[i]
      when "$handrail_depth"
        $handrail_depth =results[i]
      when "$handrail_diam"
        $handrail_diam = results[i]
      when "$panel_start_height"
        $panel_start_height = results[i]
      when "$panel_height"
        $panel_height = results[i]
      when "$panel_thickness"
        $panel_thickness = results[i]
    end
    i = i + 1
  end
end
$ss.each do |element|
  $line = element
  Balustrade.new.create_balus
end
end
# Collect components and return an array of ComponentsDefinitions names
def collect_components( prefix, upprefix )
model=Sketchup.active_model
defs = model.definitions
ret_defs = []
defs.each do |c|
  if not c.name["Image"] and not c.name["Group#"]
    if c.name[prefix] or c.name[upprefix]
      ret_defs = ret_defs.push(c.name)
    end
  end
end
return ret_defs
end
# Search the component type: 2D or 3D
def comp_type( comp )
model=Sketchup.active_model
if not model.definitions[comp]
  return nil
end
bb = model.definitions[comp].bounds
ret = nil
# depth = 0 => XY 2D component --- width = 0 => YZ 2D component --- height = 0 => XZ 2D component
d = bb.depth
w = bb.width
h = bb.height
if (tolerance d, 0.0, 0.0001) and (not (tolerance w, 0.0, 0.0001)) and (not (tolerance h, 0.0, 0.0001))
  ret = "XY"
end
if (not (tolerance d, 0.0, 0.0001)) and (tolerance w, 0.0, 0.0001) and (not (tolerance h, 0.0, 0.0001))
  ret = "YZ"
end
if (not (tolerance d, 0.0, 0.0001)) and (not (tolerance w, 0.0, 0.0001)) and (tolerance h, 0.0, 0.0001)
  ret = "XZ"
end
if (not (tolerance d, 0.0, 0.0001)) and (not (tolerance w, 0.0, 0.0001)) and (not (tolerance h, 0.0, 0.0001))
  ret = "3D"
end
return ret
end
def tolerance( n1, n2, interval)
if ( n1.between?((n2 - interval), (n2 + interval)) )
  return true
  else
  return false
end
end
# Return angle between v1 and v2, 0->2PI
def angle_between_vec( v2 )
v1 = Geom::Vector3d.new(1, 0, 0)
v2x = v2[0]
v2y = v2[1]
v2z = v2[2]
x = 1.0
y = 1.0
x = (tolerance(v2x, 0.0, 0.00001))
y = (tolerance(v2y, 0.0, 0.00001))
# 0,90,180,270
if (x == true) and ( y == false )
  if ( v2y > 0.0 ) and ( y == false )
    a = (Math::PI / 2.0)
  end
  if ( v2y < 0.0 ) and ( y == false )
    a = (3 * (Math::PI / 2.0))
  end
end
if ( x == false ) and (y == true)
  if ( v2x > 0.0 ) and (y == true)
    a = 0.0
  end
  if ( v2x < 0.0 ) and (y == true)
    a = Math::PI
  end
end
# Other angles
if ( v2x > 0.0 ) and ( y == false ) and ( x == false )
    if ( v2y > 0.0 ) and ( y == false )
      a = v1.angle_between( v2 )
    end
    if ( v2y < 0.0 ) and ( y == false )
      a = (Math::PI * 2.0) - v1.angle_between( v2 )
    end
end
if ( v2x < 0.0 ) and ( y == false ) and ( x == false )
    if ( v2y > 0.0 ) and ( y == false )
      a = v1.angle_between( v2 )
    end
    if ( v2y < 0.0 ) and ( y == false )
      a = Math::PI - v1.angle_between( v2 )
    end
end
return a
end
# Returns an array of 4 points to draw a vertical rectangle (l x w), p is the center of rectangle, v is the normal to rectangle
def vertical_section( p, v, l, w)
vx = v.axes[0]
vmx = v.axes[0].reverse
vx.length = l/2.0
vmx.length = l/2.0
vy = v.axes[1]
vmy = v.axes[1].reverse
vy.length = w/2.0
vmy.length = w/2.0
tx = Geom::Transformation.new(vx)
tmx = Geom::Transformation.new(vmx)
ty = Geom::Transformation.new(vy)
tmy = Geom::Transformation.new(vmy)
pt = p.transform tmy
p1 = pt.transform tmx
p2 = p1.transform ty
p2 = p2.transform ty
p3 = p2.transform tx
p3 = p3.transform tx
p4 = p3.transform tmy
p4 = p4.transform tmy
return [p1, p2, p3, p4]
end
# Returns an array of 4 points to draw a horizontal rectangle (l x w), p is the center of rectangle, v is the xaxis of rectangle
def horizontal_section( p, v, l, w)
#model = Sketchup.active_model
#entities = model.active_entities
vx = v
vmx = v.reverse
vx.length = l/2.0
vmx.length = l/2.0
rot =Geom::Transformation.rotation p, Geom::Vector3d.new(0, 0, 1), Math::PI/2.0
vy = vx.transform rot
vmy = vy.reverse
vy.length = w/2.0
vmy.length = w/2.0
tx = Geom::Transformation.new(vx)
tmx = Geom::Transformation.new(vmx)
ty = Geom::Transformation.new(vy)
tmy = Geom::Transformation.new(vmy)
pt = p.transform tmy
p1 = pt.transform tmx
p2 = p1.transform ty
p2 = p2.transform ty
p3 = p2.transform tx
p3 = p3.transform tx
p4 = p3.transform tmy
p4 = p4.transform tmy
return [p1, p2, p3, p4]
end
# return an array of all faces in a component
def search_faces_in_component( ents )
ret = []
  ents.each { | e | 
    if e.typename == "Face"
      ret.push(e)
    end
  }
return ret
end
def ss2array( selection_set )
ret = []
  selection_set.each { | e |
    ret.push(e)
  }
return ret
end
class Balustrade
def create_balus
model = Sketchup.active_model
entities = model.active_entities
definitions = model.definitions
start_point = $line.start.position
end_point = $line.end.position
length = $line.length
line = $line.line
vec = line[1]
model.start_operation "Create Balustrade"
entities = model.active_entities
group = entities.add_group
entities = group.entities
# Posts heights and length of panels
#-----------------------------------
# rectangular handrail, rectangular posts, rectangular panels
if $handrail_depth and $post_depth and $panel_section == "Rectangular"
  post_height = $height - $handrail_height
  length_panel = $jump - $post_width
  half_post = $post_width / 2.0
end
# rectangular handrail, rectangular posts, no panels
if $handrail_depth and $post_depth and $panel_section == "None"
  post_height = $height - $handrail_height
  half_post = $post_width / 2.0
end
# rectangular handrail, rectangular posts, component panels
if $handrail_depth and $post_depth and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - $handrail_height
  half_post = $post_width / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = model.definitions[$panel_section].bounds.width
    when "YZ"
      length_panel = $jump - $post_width
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# rectangular handrail, circular posts, rectangular panels
if $handrail_depth and $post_diam and $panel_section == "Rectangular"
  post_height = $height - $handrail_height
  length_panel = $jump - $post_diam
  half_post = $post_diam / 2.0
end
# rectangular handrail, circular posts, no panels
if $handrail_depth and $post_diam and $panel_section == "None"
  post_height = $height - $handrail_height
  half_post = $post_diam / 2.0
end
# rectangular handrail, circular posts, component panels
if $handrail_depth and $post_diam and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - $handrail_height
  half_post = $post_diam / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = model.definitions[$panel_section].bounds.width
    when "YZ"
      length_panel = $jump - $post_diam
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# rectangular handrail, component posts, rectangular panels
if $handrail_depth and !$post_depth and !$post_diam and $panel_section == "Rectangular"
  post_height = $height - $handrail_height
  half_post = model.definitions[$post_section].bounds.width / 2.0
  post_type = (comp_type $post_section)
  case post_type
    when "XY"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "YZ"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "3D"
      length_panel = $jump - model.definitions[$post_section].bounds.width
  end
end
# rectangular handrail, component posts, no panels
if $handrail_depth and !$post_depth and !$post_diam and $panel_section == "None"
  post_height = $height - $handrail_height
  half_post = model.definitions[$post_section].bounds.width / 2.0
  post_type = (comp_type $post_section)
  case post_type
    when "XY"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "YZ"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "3D"
      length_panel = $jump - model.definitions[$post_section].bounds.width
  end
end
# rectangular handrail, component posts, component panels
if $handrail_depth and !$post_depth and !$post_diam and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - $handrail_height
  half_post = model.definitions[$post_section].bounds.width / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = model.definitions[$panel_section].bounds.width
    when "YZ"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# circular handrail, rectangular posts, rectangular panels
if $handrail_diam and $post_depth and $panel_section == "Rectangular"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = $post_width / 2.0
  length_panel = $jump - $post_width
end
# circular handrail, rectangular posts, no panels
if $handrail_diam and $post_depth and $panel_section == "None"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = $post_width / 2.0
end
# circular handrail, rectangular posts, component panels
if $handrail_diam and $post_depth and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = $post_width / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = model.definitions[$panel_section].bounds.width
    when "YZ"
      length_panel = $jump - $post_width
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# circular handrail, circular posts, rectangular panels
if $handrail_diam and $post_diam and $panel_section == "Rectangular"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = $post_diam / 2.0
  length_panel = $jump - $post_diam
end
# circular handrail, circular posts, no panels
if $handrail_diam and $post_diam and $panel_section == "None"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = $post_diam / 2.0
end
# circular handrail, circular posts, component panels
if $handrail_diam and $post_diam and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = $post_diam / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = $jump - $post_diam
    when "YZ"
      length_panel = $jump - $post_diam
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# circular handrail, component posts, rectangular panels
if $handrail_diam and !$post_depth and !$post_diam and $panel_section == "Rectangular"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = model.definitions[$post_section].bounds.width / 2.0
  length_panel = $jump - model.definitions[$post_section].bounds.width
end
# circular handrail, component posts, no panels
if $handrail_diam and !$post_depth and !$post_diam and $panel_section == "None"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = model.definitions[$panel_section].bounds.width / 2.0
end
# circular handrail, component posts, component panels
if $handrail_diam and !$post_depth and !$post_diam and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - ($handrail_diam / 2.0)
  half_post = model.definitions[$post_section].bounds.width / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "YZ"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# component handrail, rectangular posts, rectangular panels
if !$handrail_depth and !$handrail_diam and $post_depth and $panel_section == "Rectangular"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = $post_width / 2.0
  length_panel = $jump - $post_width
end
# component handrail, rectangular posts, no panels
if !$handrail_depth and !$handrail_diam and $post_depth and $panel_section == "None"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = $post_width / 2.0
end
# component handrail, rectangular posts, component panels
if !$handrail_depth and !$handrail_diam and $post_depth and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = $post_width / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = $jump - $post_width
    when "YZ"
      length_panel = $jump - $post_width
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# component handrail, circular posts, rectangular panels
if !$handrail_depth and !$handrail_diam and $post_diam and $panel_section == "Rectangular"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = $post_diam / 2.0
  length_panel = $jump - $post_diam
end
# component handrail, circular posts, no panels
if !$handrail_depth and !$handrail_diam and $post_diam and $panel_section == "None"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = $post_diam / 2.0
end
# component handrail, circular posts, component panels
if !$handrail_depth and !$handrail_diam and $post_diam and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = $post_diam / 2.0
  panel_type = (comp_type $panel_section)
  case panel_type
    when "XY"
      length_panel = $jump - $post_diam
    when "YZ"
      length_panel = $jump - $post_diam
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# component handrail, component posts, rectangular panels
if !$handrail_depth and !$handrail_diam and !$post_diam and !$post_depth and $panel_section == "Rectangular"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = model.definitions[$post_section].bounds.width / 2.0
  post_type = (comp_type $post_section)
  case post_type
    when "XY"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "YZ"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# component handrail, component posts, no panels
if !$handrail_depth and !$handrail_diam and !$post_diam and !$post_depth and $panel_section == "None"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = model.definitions[$post_section].bounds.width / 2.0
end
# component handrail, component posts, component panels
if !$handrail_depth and !$handrail_diam and !$post_diam and !$post_depth and $panel_section != "Rectangular" and $panel_section != "None"
  post_height = $height - (model.definitions[$handrail_section].bounds.height / 2.0)
  half_post = model.definitions[$post_section].bounds.width / 2.0
  post_type = (comp_type $post_section)
  case post_type
    when "XY"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "YZ"
      length_panel = $jump - model.definitions[$post_section].bounds.width
    when "3D"
      length_panel = model.definitions[$panel_section].bounds.width
  end
end
# ---------------------------------------------------- Handrail creation
  Sketchup::set_status_text($exStrings.GetString("Handrail Creation"), SB_PROMPT)
  if $handrail_height
    # Rectangular section, Follow me, only once
    if ( $handrail_done == false )
      handrail_start_point = [start_point.x, start_point.y, start_point.z + $height - ($handrail_height / 2.0)]
      pts = vertical_section(handrail_start_point, vec, $handrail_depth, $handrail_height)
      handrail_face = entities.add_face pts
      ss_array = ss2array( $ss )
      handrail_face.followme ss_array
      $handrail_done = true
    end
  end
  if $handrail_diam
    # Circular section, Follow me, only once
    if ( $handrail_done == false )
      handrail_start_point = [start_point.x, start_point.y, start_point.z + $height - ($handrail_diam / 2.0)]
      section = entities.add_circle handrail_start_point, vec, ($handrail_diam/2.0), 10
      circle = entities.add_face section
      ss_array = ss2array( $ss )
      circle.reverse!.followme ss_array
      $handrail_done = true
    end
  end
if !$handrail_height and !$handrail_diam and $handrail_section != "None"
  # Component section
  # Follow me, only once
  if ( $handrail_done == false )
    handrail_start_point = [start_point.x, start_point.y, start_point.z + ($height - (model.definitions[$handrail_section].bounds.height / 2.0))]
    tt = Geom::Transformation.new handrail_start_point
    comp_start_point = handrail_start_point
    rot_angle = Geom::Vector3d.new(1, 0, 0).angle_between vec
    tr = Geom::Transformation.rotation start_point, [0,0,1], rot_angle
    handrail_component = entities.add_instance definitions[$handrail_section], tt
    handrail_component.transform! tr
    handrail_entities = handrail_component.explode
    faces = search_faces_in_component( handrail_entities )
    ss_array = ss2array( $ss )
    faces.each do |face|
      face.followme ss_array
    end
    $handrail_done = true
  end
end
# ---------------------------------------------------- Posts
   Sketchup::set_status_text($exStrings.GetString("Posts Creation"), SB_PROMPT)
  if $post_depth
    # First rectangular post
    v = vec
    v.length = $post_width / 2.0
    t = Geom::Transformation.new(v)
    post_start_point = start_point.transform t
    pts = horizontal_section(post_start_point, vec, $post_width, $post_depth)
    base = entities.add_face pts
    base.reverse!.pushpull post_height
    length_done = $post_width / 2.0
    # Intermediate rectangular posts
    if length>$jump
      while length_done < length - (($post_width * 3) / 2.0) - $jump
        vec.length = $jump
        tj = Geom::Transformation.new(vec)
        post_start_point.transform! tj
        pts = horizontal_section(post_start_point, vec, $post_width, $post_depth)
        base = entities.add_face pts
        base.reverse!.pushpull post_height
        length_done = length_done + $jump
      end
    end
    # Last rectangular post
    v = vec.reverse
    v.length = $post_width / 2.0
    t = Geom::Transformation.new(v)
    post_start_point = end_point.transform t
    pts = horizontal_section(post_start_point, vec, $post_width, $post_depth)
    base = entities.add_face pts
    base.reverse!.pushpull post_height
    length_done = $post_width / 2.0
  end
  
  if $post_diam
  # First circular post
    v = vec
    v.length = $post_diam / 2.0
    t = Geom::Transformation.new(v)
    post_start_point = start_point.transform t
    section = entities.add_circle post_start_point, vec.axes[1], ($post_diam/2.0), 10
    circle = entities.add_face section
    if( circle.normal.dot(Z_AXIS) > 0 )
      circle.pushpull post_height
      else
      circle.reverse!.pushpull post_height
    end
    length_done = $post_diam / 2.0
      
  # Intermediate circular posts
    if length>$jump
      while length_done < length - (($post_diam * 3) / 2.0) - $jump
        vec.length = $jump
        tj = Geom::Transformation.new(vec)
        post_start_point.transform! tj
        section = entities.add_circle post_start_point, vec.axes[1], ($post_diam/2.0), 10
        circle = entities.add_face section
        if( circle.normal.dot(Z_AXIS) > 0 )
          circle.pushpull $height-($post_diam/2.0)
          else
          circle.reverse!.pushpull post_height
        end
        length_done = length_done + $jump
      end
    end    
  # Last circular post
    v = vec.reverse
    v.length = $post_diam / 2.0
    t = Geom::Transformation.new(v)
    post_start_point = end_point.transform t
    section = entities.add_circle post_start_point, vec.axes[1], ($post_diam/2.0), 10
    circle = entities.add_face section
    if( circle.normal.dot(Z_AXIS) > 0 )
      circle.pushpull $height-($post_diam/2.0)
      else
      circle.reverse!.pushpull post_height
    end
    length_done = $post_diam / 2.0
  end
  
  if ($post_section != "Rectangular") and ($post_section != "Circular")
    # is the component 2D or 3D ?
    post_type = (comp_type $post_section)
    v = vec
    v.length = half_post
    #First component post
    comp_start_point = start_point.transform! Geom::Transformation.translation(v)
    tt = Geom::Transformation.new comp_start_point
    rot_angle = angle_between_vec vec
    tr = Geom::Transformation.rotation comp_start_point, [0,0,1], rot_angle
    post_component = entities.add_instance definitions[$post_section], tt
    post_component.transform! tr
    length_done = half_post
    # Explode and extrude the component only if its type is "XY"
    if (post_type == "XY")
      post_entities = post_component.explode
      faces = search_faces_in_component( post_entities )
        faces.each do |face|
          face.pushpull post_height
        end
    end
    
    # Intermediate component posts
    if length>$jump
      while length_done < length - $jump
        v = vec
        v.length = $jump
        comp_start_point = comp_start_point.offset v
        tt = Geom::Transformation.new comp_start_point
        rot_angle = angle_between_vec vec
        tr = Geom::Transformation.rotation comp_start_point, [0,0,1], rot_angle
        post_component = entities.add_instance definitions[$post_section], tt
        post_component.transform! tr
        # Explode and extrude the component only if its type is "XY"
        if (post_type == "XY")
        post_entities = post_component.explode
        faces = search_faces_in_component( post_entities )
          faces.each do |face|
            face.pushpull post_height
          end
        end
        length_done = length_done + $jump
      end
    end
    
    #Last component post
    v = vec
    v.length = length - (half_post * 2.0)
    comp_start_point = start_point.offset v
    tt = Geom::Transformation.new comp_start_point
    rot_angle = angle_between_vec vec
    tr = Geom::Transformation.rotation comp_start_point, [0,0,1], rot_angle
    post_component = entities.add_instance definitions[$post_section], tt
    post_component.transform! tr
    # Explode and extrude the component only if its type is "XY"
    if (post_type == "XY")
      post_entities = post_component.explode
      faces = search_faces_in_component( post_entities )
        faces.each do |face|
          face.pushpull post_height
        end
    end
  end # of component posts
# -------------------------------------------- Rectangular panels
   Sketchup::set_status_text($exStrings.GetString("panels Creation"), SB_PROMPT)
if ($panel_section == "Rectangular")
  num_panels = (length/$jump).floor.to_i
  #puts num_panels.to_s + "\n"
  v1 = vec
  length_done = 0.0
  
  if $post_width and $post_depth
    # Rect posts
    v1.length = ($post_width / 2.0) + ($jump / 2.0)
  end
  if $post_diam and !$post_width and !$post_depth
    # Circ posts
    v1.length = ($post_diam / 2.0) + ($jump / 2.0)
  end
  if !$post_diam and !$post_width and !$post_depth
    # Component posts
    v1.length = ($jump / 2.0)
  end
  
  t1 = Geom::Transformation.translation(v1)
  tup = Geom::Transformation.new Geom::Vector3d.new(0, 0, $panel_start_height) 
  
  # First panel
  panel_start_point = start_point.transform t1
  panel_start_point = panel_start_point.transform tup
  pts = horizontal_section(panel_start_point, vec, length_panel, $panel_thickness)
  panel = entities.add_face pts
  if( panel.normal.dot(Z_AXIS) > 0 )
    panel.pushpull $panel_height
  else
    panel.reverse!.pushpull $panel_height
  end
  length_done = ($jump / 2.0)
  
  # Other panels
  vt = vec
  vt.length = $jump
  t2 = Geom::Transformation.translation(vt)
  0.upto (num_panels - 2) do |i|
    panel_start_point= panel_start_point.transform t2
    pts = horizontal_section(panel_start_point, vec, length_panel, $panel_thickness)
    panel = entities.add_face pts
    if( panel.normal.dot(Z_AXIS) > 0 )
      panel.pushpull $panel_height
    else
      panel.reverse!.pushpull $panel_height
    end
    length_done = length_done + $jump
  end
  
  # Last panel
  #puts length_done.to_cm.to_s + "\n"
  l_rest = (length - length_done) - (($jump / 2.0) + (half_post * 4))
  if l_rest > 0
    vtt = vec
    length_panel = (length - length_done) - (($jump / 2.0) + (half_post * 4))
    l_rest = (length - length_done) - (($jump / 2.0) + (half_post * 4))
    #puts length_panel.to_cm.to_s + "\n"
    vtt.length = ($jump / 2.0) + half_post + (length_panel / 2.0) 
    t3 = Geom::Transformation.translation(vtt)
    panel_start_point= panel_start_point.transform t3
    pts = horizontal_section(panel_start_point, vec, length_panel, $panel_thickness)
    panel = entities.add_face pts
    if( panel.normal.dot(Z_AXIS) > 0 )
      panel.pushpull $panel_height
    else  
      panel.reverse!.pushpull $panel_height
    end
  end
  length_done = length_done + $jump
end #draw rectangular panels
# -------------------------------------------- Component panels
if ($panel_section != "Rectangular") and ($panel_section != "None")
  num_panels = (length/$jump).floor.to_i
  v1 = vec
  length_done = 0.cm
  panel_type = (comp_type $panel_section)
  
  if $post_width and $post_depth
    # Rect posts
    v1.length = ($post_width / 2.0) + ($jump / 2.0)
  end
  if $post_diam and !$post_width and !$post_depth
    # Circ posts
    v1.length = ($post_diam / 2.0) + ($jump / 2.0)
    #half_post = $post_diam / 2.0
  end
  if !$post_diam and !$post_width and !$post_depth
    # Component
    v1.length = ($jump / 2.0)
  end
  
  tup = Geom::Transformation.new Geom::Vector3d.new(0, 0, $panel_start_height) 
  tjump = Geom::Transformation.translation(v1)
  
  comp_start_point = start_point.transform tup
  comp_start_point.transform! tjump
 
 # First component panel
  tt = Geom::Transformation.translation(comp_start_point)
  rot = angle_between_vec vec
  tr = Geom::Transformation.rotation comp_start_point, [0,0,1], rot
  panel_component = entities.add_instance definitions[$panel_section], tt
  panel_component.transform! tr
  # Explode and extrude the component only if its type is "XY"
    if (panel_type == "XY")
      panel_entities = panel_component.explode
      faces = search_faces_in_component( panel_entities )
        faces.each do |face|
          face.pushpull $panel_height
        end
    end
    # Explode and extrude the component only if its type is "YZ"
    if (panel_type == "YZ")
      panel_entities = panel_component.explode
      faces = search_faces_in_component( panel_entities )
        faces.each do |face|
          face.pushpull (0.0 - (length_panel / 2.0))
          face.pushpull (length_panel / 2.0)
        end
    end
  length_done = $jump / 2.0
  
# Other components panels
  vt = vec
  vt.length = $jump
  tjump = Geom::Transformation.translation(vt)
  #puts num_panels.to_s + "\n"
  0.upto (num_panels - 2) do |i|
    comp_start_point = comp_start_point.transform! tjump
    tt = Geom::Transformation.translation(comp_start_point)
    panel_component = entities.add_instance definitions[$panel_section], tt
    rot = angle_between_vec vec
    tr = Geom::Transformation.rotation comp_start_point, [0,0,1], rot
    panel_component.transform! tr
    # Explode and extrude the component only if its type is "XY"
    if (panel_type == "XY")
      panel_entities = panel_component.explode
      faces = search_faces_in_component( panel_entities )
        faces.each do |face|
          face.pushpull $panel_height
        end
    end
    # Explode and extrude the component only if its type is "YZ"
    if (panel_type == "YZ")
      panel_entities = panel_component.explode
      faces = search_faces_in_component( panel_entities )
        faces.each do |face|
          face.pushpull (0.0 - (length_panel / 2.0))
          face.pushpull (length_panel / 2.0)
        end
    end
    length_done = length_done + $jump
  end
  
  # Last component panel
  vt = vec
  l_rest = (length - length_done) - (($jump / 2.0) + (half_post * 4))
  #puts l_rest.to_cm.to_s + "\n"
  if (panel_type == "XY") or (panel_type == "3D")
    vt.length = $jump
    else
    vt.length = ($jump / 2.0) + half_post
  end
  if l_rest > 0
    tjump = Geom::Transformation.translation(vt)
    length_panel = (length - length_done) - ($jump / 2.0) - (half_post * 4.0)  
    comp_start_point = comp_start_point.transform! tjump
    tt = Geom::Transformation.translation(comp_start_point)
    
    if (panel_type == "YZ")
    panel_component = entities.add_instance definitions[$panel_section], tt
    rot = angle_between_vec vec
    tr = Geom::Transformation.rotation comp_start_point, [0,0,1], rot
    panel_component.transform! tr
    end
    # Explode and extrude the component only if its type is "YZ"
    if (panel_type == "YZ")
      panel_entities = panel_component.explode
      faces = search_faces_in_component( panel_entities )
        faces.each do |face|
          face.pushpull -length_panel
        end
    end
  end
end #draw components panels
  model.commit_operation
  Sketchup::set_status_text($exStrings.GetString("Balustrade Creation done."), SB_PROMPT)
end 
end #class Balustrade
#***********************************************************************************************************************************************************************************
#                       SKETCHY FFD
#***********************************************************************************************************************************************************************************
#SketchyFFD
#Copyright 2008 Chris Phillips
#
Infinity = 1.0/0
#Right click UI menu.
#UI.add_context_menu_handler { |menu|
#    menu=menu.add_submenu("- FFD")
#    if(!Sketchup.active_model.selection.empty? && Sketchup.active_model.selection[0].typename=="Group")
#        menu.add_item("2x2 FFD"){startFFD(Sketchup.active_model.selection[0],[2,2,2],"false")}
#        menu.add_item("3x3 FFD"){startFFD(Sketchup.active_model.selection[0],[3,3,3],"false")}
#        menu.add_item("NxN FFD"){
#            prompts = ["Width","Depth","Height","Subdivide"]
#            values = [4, 4, 4,"false"]
#            results = inputbox prompts, values,["","","","true|false"], "FFD Dimensions"
#            if (results)
#                startFFD(Sketchup.active_model.selection[0],results,results[3]) 
#            end
#        }
#    end
#	menu.add_item("Lock edges"){
#			Sketchup.active_model.start_operation "Lock edges"
#		Sketchup.active_model.selection.each{|ent|
#		    if(ent.class==Sketchup::Edge)
#			ent.vertices.each{|v| v.set_attribute("SFFD","locked",true)}
#		    end
#		}
#		Sketchup.active_model.commit_operation
#	}    
#	menu.add_item("Unlock edges"){
#		Sketchup.active_model.start_operation "Unlock edges"
#		Sketchup.active_model.selection.each{|ent|
#		    if(ent.class==Sketchup::Edge)
#			ent.vertices.each{|v| v.set_attribute("SFFD","locked",false)}
#		    end
#		}
#		Sketchup.active_model.commit_operation
#	}   
#    menu.add_item("Make patch..."){
#        prompts = ["Width","Depth","Cell Width","Cell Depth"]
#        #prompts = []
#        values = [4, 4,10,10]
#       results = inputbox prompts, values, "Patch FFD dimensions"
#        
#        Sketchup.active_model.start_operation("Create FFD patch")
#        grp=Sketchup.active_model.entities.add_group()
#        0.upto(results[0]-2){|w|
#            0.upto(results[1]-2){|h|
#                xform=Geom::Transformation.new([w*results[2],h*results[3],0])
#                xform=xform* Geom::Transformation.scaling(results[2],results[3], 1.0)
#                f=[[0,0,0].transform(xform),
#                    [0,1,0].transform(xform),
#                    [1,1,0].transform(xform),
#                    [1,0,0].transform(xform)]
#                grp.entities.add_face(f)
#            }
#        }
#        Sketchup.active_model.commit_operation
#        Sketchup.active_model.selection.clear
#        Sketchup.active_model.selection.add(grp)
#        if (results)
#            startFFD(Sketchup.active_model.selection[0],[results[0],results[1],1]) 
#        end
#    }
#}
#This class provides events when the user manipulates the control points
#NOTE. The events work but the hooks have been commented out. 
#Search for observer to enable these events.
class LatticeObserver
    def onElementAdded(entity, xx)
    end
 	def onContentsModified (entity)
	end
        #called when lattice point is moved.
 	def onChangeEntity(entity)
        #puts "changed "+entity.to_s
        if(entity.class==Sketchup::ComponentDefinition)
            entity.set_attribute("controlLattice","currentTransformation",entity.transformation.to_a)    
        elsif(entity.class==Sketchup::ConstructionPoint && $latticeGroup!=nil)
            updateFFD
        end
        
        #if(entity.get_attribute("controlLattice","currentTransformation",nil)!=nil)
            #puts entity.get_attribute("controlLattice","currentTransformation",nil)
        #end
        #~ parent= Sketchup.active_model.active_entities.parent
        #~ if parent.class==Sketchup::ComponentDefinition
           #~ if parent.group?
               #~ trans=parent.instances[0].transformation
               #~ p "in group #{parent.name}"
               #~ p  trans.origin
           #~ end
        #~ end
        
    end
        #called when lattice group is opened
    def onOpen(instance)
        instance.set_attribute("controlLattice","isOpen",true)
    end
        #called when lattice group is closed
    def onClose(instance)
        instance.set_attribute("controlLattice","isOpen",false)
    end
end    
#Create a control lattice and calculate the vertex weights for a group.
def startFFD(grp,numControlPoints,bSubdivide)
    $ffdGroup=grp
	dice_group(grp,[numControlPoints[0]-1,numControlPoints[1]-1,numControlPoints[2]-1]) if(bSubdivide=="true")
	
    #convert and sanity check.
    numControlPoints[0]=(numControlPoints[0]-1).to_f
    numControlPoints[1]=(numControlPoints[1]-1).to_f
    numControlPoints[2]=(numControlPoints[2]-1).to_f
    
    size=(grp.bounds.max-grp.bounds.min).to_a#might want to get definition instead.
    if(size[0]<=0.0)
         #numControlPoints[0]=0.0
         size[0]=0.0001
    end
    if(size[1]<=0.0)
         #numControlPoints[1]=0.0
         size[1]=0.0001
    end
    if(size[2]<=0.0)
         #numControlPoints[2]=0.0
         size[2]=0.0001
    end
        #remove previous lattice (if any)
    if($latticeGroup!=nil)
        begin
            $latticeGroup.erase!
        rescue
            #group was already deleted.
        end
    end
    
        #create lattice
    $latticeGroup=createControlLattice(grp,size,numControlPoints)
    
    Sketchup.active_model.start_operation "Start FFD"
    initFFD(Sketchup.active_model.selection[0],size,numControlPoints)   
    Sketchup.active_model.commit_operation
end
#deform the group based on the position changes of the control points.
def updateFFD
    if($latticeGroup==nil || $latticeGroup.deleted?)
        $latticeGroup==nil
        return
    end
    
    Sketchup.active_model.start_operation "FFD"
    analizeLattice($latticeGroup,$ffdGroup)
    Sketchup.active_model.commit_operation
end
    
#called from startFFD to calculate the groups vertex weights.    
def initFFD(grp,size,numControlPoints)
    #size=grp.bounds.max-grp.bounds.min
        #make sure this group is unique or chaos insues.    
    grp.make_unique
        #global. needs to be made a class varible somehow.
    $allVerts=[]
    grp.entities.each{|ent|
    puts ent
        if(ent.typename == "Edge" && ent.curve!=nil)
            ent.explode_curve #all curves need to be exploded for deform to work right.
        end
        }
    grp.entities.each{|ent|
        begin#guard against entities that dont have verts.
            ent.vertices.each{|vert|
                $allVerts.push(vert)
            }
        rescue
            puts "Warning:entities of type #{ent.typename} can't be deformed."
        end
    }
        #allverts now contains redundant verts. remove duplicates.
    $allVerts.uniq!
    #puts $allVerts.length
    
        #Global to hold the calculated weight for each vertex per control point.
    $allVertWeights=[]
        #this loop could be dramaticly optimized.
    vi=0        
    $allVerts.each{|vert|
        stuv=vert.position
        stuv.x=stuv.x/size[0];stuv.y=stuv.y/size[1];stuv.z=stuv.z/size[2]
            #calc weights.
        stuv.x=1.0 if(stuv.x==Infinity)    
        stuv.y=1.0 if(stuv.y==Infinity)    
        stuv.z=1.0 if(stuv.z==Infinity)    
#puts stuv.inspect            
        weights=[]
        0.upto(numControlPoints[0]){|x|
            bx=calcBernstein(x,numControlPoints[0],stuv.x)
            0.upto(numControlPoints[1]){|y|
                bxy=bx*calcBernstein(y,numControlPoints[1],stuv.y)
                0.upto(numControlPoints[2]){|z|
                    weights.push(bxy * calcBernstein(z,numControlPoints[2],stuv.z))
                }
            }
        }
        $allVertWeights.push(weights)
	vert.set_attribute("SFFD","locked",false)
        vi=vi+1
        Sketchup.set_status_text ("Weighing  #{vi} of #{$allVerts.length}") if(vi%100==0)
    }
end
#deform the mesh based on the change in position of each control point.
#deltas is an array. 
#each element in the deltas array is an array with 2 elements. 
#the first element is the index of the control point.
#the second element is the change in postition of that point. (a array of 3 floats).
#example:
#deltas=[]
#deltas.push([0,[0.0,0.0,1.0]])#move control point 0 "up" 1.0
#deltas.push([5,[0.0,0.0,-2.0]])#move control point 5 "down" 2.0
#etc
def applyMultipleFFD(grp,deltas)
    
    #grp.entities.transform_entities(Geom::Transformation.new([0.0,0.0,0.001]),grp.entities.to_a)
    vi=0;#used to loop through $allVertWeights in the same order as $allVerts
    $allVerts.each{|vert|
	
        weights=$allVertWeights[vi] #get pre-calculated array of vertex weights (one per control point)
        vi=vi+1
	
	next if(vert.get_attribute("SFFD","locked",false))
            #accumulate the change in position of this vertex based on 
            #each control points weighted movement.
        dvect=[0.0,0.0,0.0] 
        deltas.each{|delta|
            weight=weights[delta[0]] #delta[0] = control point index.
#puts weight            
            vect=delta[1]#delta[1]=delta vector for control point
            dvect[0]+=vect[0]*weight
            dvect[1]+=vect[1]*weight
            dvect[2]+=vect[2]*weight
        }
        #move the vertex.
        begin
        grp.entities.transform_entities(Geom::Transformation.new(dvect),vert)
        rescue
        end
        #display progress
        Sketchup.set_status_text ("Deforming #{vi} of #{$allVerts.length}") if(vi%100==0)
        #puts("Deforming #{vi} of #{$allVerts.length}") if(vi%100==0)
    }
end
#deform by a single moved control point
#not used anymore since applyMultipleFFD is much faster for multiple moved cp's
def applyFFD(grp,index,vect)
    vi=0;
    $allVerts.each{|vert|
        weights=$allVertWeights[vi] ##weights=vert.get_attribute("FFD","weights",nil)
        weight=weights[index]
        grp.entities.transform_entities(Geom::Transformation.new([vect[0]*weight,vect[1]*weight,vect[2]*weight]),vert)
        vi=vi+1
    }
end
#helper function to create an array containing 
#the position of each control point based 
#on the size of the bounding box.
def calculateControlLattice(size,numControlPoints)
    #puts numControlPoints
    xstep=size[0].to_f/numControlPoints[0]
    ystep=size[1].to_f/numControlPoints[1]    
    zstep=size[2].to_f/numControlPoints[2]
    #handle dim of 0 (1)
    xstep=1.0 if(xstep==Infinity)    
    ystep=1.0 if(ystep==Infinity)    
    zstep=1.0 if(zstep==Infinity)    
                    
    controlPoints=[]
    0.upto(numControlPoints[0]){|x|
        0.upto(numControlPoints[1]){|y|
            0.upto(numControlPoints[2]){|z|
                #NOTE: 0.001+ is because bug in sketchup will not allow construction point at 0,0,0
                controlPoints.push([0.001+(x*xstep),0.001+(y*ystep),0.001+(z*zstep)])
            }
        }
    }
    return(controlPoints)
end
def dice_group(grp,numControlPoints)
	Sketchup.active_model.start_operation("Dice")
	
	diceGroup=createDicePlanes(grp,numControlPoints)
	grp.entities.intersect_with(false,grp.transformation,grp.entities,diceGroup.transformation,false,[diceGroup])
	diceGroup.erase!
	
	Sketchup.active_model.commit_operation
end
def createDicePlanes(grp,numControlPoints)
	origin=grp.bounds.min
	size=(grp.bounds.max-grp.bounds.min).to_a
	diceGroup=Sketchup.active_model.active_entities.add_group
	
	xstep=size[0].to_f/numControlPoints[0]
	ystep=size[1].to_f/numControlPoints[1]    
	zstep=size[2].to_f/numControlPoints[2]
	
	#handle dim of 0 (1)
	xstep=1.0 if(xstep==Infinity)    
	ystep=1.0 if(ystep==Infinity)    
	zstep=1.0 if(zstep==Infinity)    
                    
    
	1.upto(numControlPoints[0]-1){|x|
		diceGroup.entities.add_face([[x*xstep,0,0],[x*xstep,0,size[2]],[x*xstep,size[1],size[2]],[x*xstep,size[1],0]])
	}
	1.upto(numControlPoints[1]-1){|y|
		diceGroup.entities.add_face([[0,y*ystep,0],[0,y*ystep,size[2]],[size[0],y*ystep,size[2]],[size[0],y*ystep,0]])
	}
	1.upto(numControlPoints[2]-1){|z|
		diceGroup.entities.add_face([[0,0,z*zstep],[0,size[1],z*zstep],[size[0],size[1],z*zstep],[size[0],0,z*zstep]])
	}
	diceGroup.transformation=grp.transformation
	return diceGroup
end
#create a group to act as a control lattice.
def createControlLattice(grp,size, numControlPoints)
    #calc min and max bounding
    #size=grp.bounds.max-grp.bounds.min
    origin=grp.bounds.min
    ents=grp.entities
    
    controlPoints=calculateControlLattice(size,numControlPoints)
    Sketchup.active_model.start_operation("create control lattice")
    latticeGroup=Sketchup.active_model.active_entities.add_group
        #create control lattice group.
    entities=latticeGroup.entities
        #create an observer for events. Not used yet.
    observer=LatticeObserver.new
    
    #create construction points.
    index=0
    controlPoints.each{|lcp|
        cpt=entities.add_cpoint(Geom::Point3d.new([lcp[0],lcp[1],lcp[2]]))
        
        cpt.set_attribute("controlPoint","originalPosition",cpt.position.to_a)
        cpt.set_attribute("controlPoint","index",index)
        index=index+1
        
        #trigger event any time the user moves a point. not used yet.
        cpt.add_observer(observer) 
    }
    #move lattice group to match ffd group
    latticeGroup.transformation=grp.transformation
    latticeGroup.set_attribute("controlLattice","currentTransformation",latticeGroup.transformation.to_a)
    
    #monitor group open/close event. not used yet.
    latticeGroup.add_observer(observer)
    
    Sketchup.active_model.commit_operation
    return(latticeGroup)
end
#determine which (if any) control points have been moved and deform the verts in the ffd group.
def analizeLattice(latticeGroup,ffdGroup)
    
    xform=Geom::Transformation.new()
    if(latticeGroup.get_attribute("controlLattice","isOpen",false))
        #if group is open transform by inverse of saved xform (the real xform).
        xform=Geom::Transformation.new(latticeGroup.get_attribute("controlLattice","currentTransformation",nil))
        xform.invert!
    end
    #puts entity.get_attribute("controlLattice","currentTransformation",nil)
    
    deltas=[]
    latticeGroup.entities[0].parent.entities.each{|cpt|
        op=cpt.get_attribute("controlPoint","originalPosition",nil)
        if(op!=nil)#ent is a control point.
            cp=cpt.position.transform(xform).to_a#get current position
            if(cp!=op)#point moved?
                delta=[]#calculate change
                delta[0]=cp[0]-op[0]
                delta[1]=cp[1]-op[1]
                delta[2]=cp[2]-op[2]
                cpt.set_attribute("controlPoint","originalPosition",cp)#update point so user can tweak after inital ffd.
                index=cpt.get_attribute("controlPoint","index",nil)
                deltas.push([index,delta])#add to array of changes.
            end
        end
    }
#puts deltas.inspect
    #update mesh    
    applyMultipleFFD(ffdGroup,deltas) if (deltas.length>0)
end
#used by calcBernstein. could be replaced by a table since in 
#this application n is never more than 3 (number of control points).
def factorial(n)
    sum = 1
    sum.upto(n) { |i| sum *= i }
    sum
end
#TODO:For this application this function could be replaced by a lookup since
#i=0 to number of control points (<3),n=number of control points (never changes)
#u=float that cant be predicted.
def calcBernstein (i, n, u) #Bernstein Polynomial
    binomial = factorial(n).to_f / (factorial(n - i).to_f * factorial(i).to_f)
    bernstein = binomial * (u**i) * ((1-u)**(n-i))
    return(bernstein)
end
#end
     
